from decimal import Decimal
import numpy as np
#from numpy.typing import NDArray
from typing import Union, Generic, TypeVar, Tuple, List, Dict, Callable, Any, cast
from unyt import unyt_quantity, unyt_array

from ._cast import Cast, non_nullable_cast, nullable_cast, NestedCast

# Typeparam definition
T = TypeVar("T")
K = TypeVar("K")
V = TypeVar("V")

# Some common type casts

# Non-nullables

cast_bool    : Cast[bool]    = non_nullable_cast(bool)
cast_int     : Cast[int]     = non_nullable_cast(int)
cast_float   : Cast[float]   = non_nullable_cast(float)
cast_Decimal : Cast[Decimal] = non_nullable_cast(Decimal)
cast_str     : Cast[str]     = non_nullable_cast(str)

cast_list  : Cast[List[Any]]  = non_nullable_cast(list)
cast_tuple : Cast[Tuple[Any, ...]] = non_nullable_cast(tuple)

def cast_typed_list(*type_casts: Cast[T]) -> NestedCast[List[T]]:
    return NestedCast[List[T]](cast_list, *type_casts, nullable = False)

def cast_typed_tuple(*type_casts: Cast[T]) -> NestedCast[Tuple[T]]:
    return NestedCast[Tuple[T]](cast_tuple, *type_casts, nullable = False)

def cast_dict(cast_keys: Union[Cast[K], None] = None, cast_values: Union[Cast[V], None] = None) -> Cast[Dict[K, V]]:
    if cast_keys is None:
        cast_keys = Cast.passthrough()
    if cast_values is None:
        cast_values = Cast.passthrough()
    def cast_dict_obj(value: Dict[Any, Any]) -> Dict[K, V]:
        return { cast(K, cast_keys(key)): cast(V, cast_values(value[key])) for key in value }
    return non_nullable_cast(cast_dict_obj)

def cast_ndarray(*args: Any, **kwargs: Any) -> Cast[np.ndarray]:
    """
    *args and **kwargs for numpy.array() may be passed
    """
    return non_nullable_cast(np.array, *args, **kwargs)

cast_ndarray_bool     : Cast[np.ndarray] = cast_ndarray(dtype = bool)
cast_ndarray_int      : Cast[np.ndarray] = cast_ndarray(dtype = int)
cast_ndarray_int8     : Cast[np.ndarray] = cast_ndarray(dtype = np.int8)
cast_ndarray_int16    : Cast[np.ndarray] = cast_ndarray(dtype = np.int16)
cast_ndarray_int32    : Cast[np.ndarray] = cast_ndarray(dtype = np.int32)
cast_ndarray_int64    : Cast[np.ndarray] = cast_ndarray(dtype = np.int64)
#cast_ndarray_int128   : Cast[np.ndarray] = cast_ndarray(dtype = np.int128)
#cast_ndarray_int256   : Cast[np.ndarray] = cast_ndarray(dtype = np.int256)
cast_ndarray_float    : Cast[np.ndarray] = cast_ndarray(dtype = float)
cast_ndarray_float16  : Cast[np.ndarray] = cast_ndarray(dtype = np.float16)
cast_ndarray_float32  : Cast[np.ndarray] = cast_ndarray(dtype = np.float32)
cast_ndarray_float64  : Cast[np.ndarray] = cast_ndarray(dtype = np.float64)
#cast_ndarray_float80  : Cast[np.ndarray] = cast_ndarray(dtype = np.float80)
#cast_ndarray_float96  : Cast[np.ndarray] = cast_ndarray(dtype = np.float96)
#cast_ndarray_float128 : Cast[np.ndarray] = cast_ndarray(dtype = np.float128)
#cast_ndarray_float256 : Cast[np.ndarray] = cast_ndarray(dtype = np.float256)
cast_ndarray_str      : Cast[np.ndarray] = cast_ndarray(dtype = str)

def cast_unyt_quantity(unit: str, *args: Any, **kwargs: Any) -> Cast[unyt_quantity]:
    """
    *args and **kwargs for unyt.unyt_quantity() may be passed
    (though will only be used if the target value is not already a unyt_quantity object!)
    """
    def inner(value: Any, _unit: str, *args: Any, **kwargs: Any):
        if isinstance(value, unyt_quantity):
            return value.to(_unit)
        elif isinstance(value, unyt_array):
            return unyt_quantity(value.to(_unit), _unit, *args, **kwargs)
        else:
            return unyt_quantity(value, _unit, *args, **kwargs)
    return non_nullable_cast(inner, unit, *args, **kwargs)

cast_unyt_quantity_mass_si           : Cast[unyt_quantity] = cast_unyt_quantity(unit = "kg",      dtype = np.float64)
cast_unyt_quantity_mass_imperial     : Cast[unyt_quantity] = cast_unyt_quantity(unit = "lb",      dtype = np.float64)
cast_unyt_quantity_mass_cgs          : Cast[unyt_quantity] = cast_unyt_quantity(unit = "g",       dtype = np.float64)
cast_unyt_quantity_mass_solar        : Cast[unyt_quantity] = cast_unyt_quantity(unit = "Msun",    dtype = np.float64)
cast_unyt_quantity_time_milliseconds : Cast[unyt_quantity] = cast_unyt_quantity(unit = "ms",      dtype = np.float64)
cast_unyt_quantity_time_seconds      : Cast[unyt_quantity] = cast_unyt_quantity(unit = "s",       dtype = np.float64)
cast_unyt_quantity_time_minutes      : Cast[unyt_quantity] = cast_unyt_quantity(unit = "min",     dtype = np.float64)
cast_unyt_quantity_time_hours        : Cast[unyt_quantity] = cast_unyt_quantity(unit = "hr",      dtype = np.float64)
cast_unyt_quantity_time_days         : Cast[unyt_quantity] = cast_unyt_quantity(unit = "day",     dtype = np.float64)
cast_unyt_quantity_time_years        : Cast[unyt_quantity] = cast_unyt_quantity(unit = "yr",      dtype = np.float64)
cast_unyt_quantity_time_Gyrs         : Cast[unyt_quantity] = cast_unyt_quantity(unit = "Gyr",     dtype = np.float64)
cast_unyt_quantity_length_si         : Cast[unyt_quantity] = cast_unyt_quantity(unit = "m",       dtype = np.float64)
cast_unyt_quantity_length_imperial   : Cast[unyt_quantity] = cast_unyt_quantity(unit = "ft",      dtype = np.float64)
cast_unyt_quantity_length_cgs        : Cast[unyt_quantity] = cast_unyt_quantity(unit = "cm",      dtype = np.float64)
cast_unyt_quantity_length_solar      : Cast[unyt_quantity] = cast_unyt_quantity(unit = "Rsun",    dtype = np.float64)
cast_unyt_quantity_length_pc         : Cast[unyt_quantity] = cast_unyt_quantity(unit = "pc",      dtype = np.float64)
cast_unyt_quantity_length_kpc        : Cast[unyt_quantity] = cast_unyt_quantity(unit = "kpc",     dtype = np.float64)
cast_unyt_quantity_length_Mpc        : Cast[unyt_quantity] = cast_unyt_quantity(unit = "Mpc",     dtype = np.float64)
cast_unyt_quantity_speed_si          : Cast[unyt_quantity] = cast_unyt_quantity(unit = "m/s",     dtype = np.float64)
cast_unyt_quantity_speed_imperial    : Cast[unyt_quantity] = cast_unyt_quantity(unit = "ft/s",    dtype = np.float64)
cast_unyt_quantity_speed_cgs         : Cast[unyt_quantity] = cast_unyt_quantity(unit = "cm/s",    dtype = np.float64)
cast_unyt_quantity_speed_kms         : Cast[unyt_quantity] = cast_unyt_quantity(unit = "km/s",    dtype = np.float64)
cast_unyt_quantity_speed_Mpc_per_Gyr : Cast[unyt_quantity] = cast_unyt_quantity(unit = "Mpc/Gyr", dtype = np.float64)
cast_unyt_quantity_force_si          : Cast[unyt_quantity] = cast_unyt_quantity(unit = "N",       dtype = np.float64)
cast_unyt_quantity_force_imperial    : Cast[unyt_quantity] = cast_unyt_quantity(unit = "lbf",     dtype = np.float64)
cast_unyt_quantity_force_cgs         : Cast[unyt_quantity] = cast_unyt_quantity(unit = "dyn",     dtype = np.float64)
cast_unyt_quantity_energy_si         : Cast[unyt_quantity] = cast_unyt_quantity(unit = "J",       dtype = np.float64)
cast_unyt_quantity_energy_imperial   : Cast[unyt_quantity] = cast_unyt_quantity(unit = "BTU",     dtype = np.float64)
cast_unyt_quantity_energy_cgs        : Cast[unyt_quantity] = cast_unyt_quantity(unit = "erg",     dtype = np.float64)
cast_unyt_quantity_energy_eV         : Cast[unyt_quantity] = cast_unyt_quantity(unit = "eV",      dtype = np.float64)
cast_unyt_quantity_energy_KeV        : Cast[unyt_quantity] = cast_unyt_quantity(unit = "KeV",     dtype = np.float64)
cast_unyt_quantity_energy_GeV        : Cast[unyt_quantity] = cast_unyt_quantity(unit = "GeV",     dtype = np.float64)
cast_unyt_quantity_energy_TeV        : Cast[unyt_quantity] = cast_unyt_quantity(unit = "TeV",     dtype = np.float64)
cast_unyt_quantity_power_si          : Cast[unyt_quantity] = cast_unyt_quantity(unit = "W",       dtype = np.float64)
cast_unyt_quantity_power_imperial    : Cast[unyt_quantity] = cast_unyt_quantity(unit = "hp",      dtype = np.float64)
cast_unyt_quantity_power_cgs         : Cast[unyt_quantity] = cast_unyt_quantity(unit = "erg/s",   dtype = np.float64)
cast_unyt_quantity_power_solar       : Cast[unyt_quantity] = cast_unyt_quantity(unit = "Lsun",    dtype = np.float64)

def cast_unyt_array(unit: str, *args: Any, **kwargs: Any) -> Cast[unyt_array]:
    """
    *args and **kwargs for unyt.unyt_array() may be passed
    (though will only be used if the target value is not already a unyt_array object!)
    """
    def inner(value: Any, _unit: str, *args: Any, **kwargs: Any):
        if isinstance(value, unyt_quantity):
            return unyt_array(value.to(_unit), _unit, *args, **kwargs)
        elif isinstance(value, unyt_array):
            return value.to(_unit)
        else:
            return unyt_array(value, _unit, *args, **kwargs)
    return non_nullable_cast(inner, unit, *args, **kwargs)

cast_unyt_array_mass_si           : Cast[unyt_array] = cast_unyt_array(unit = "kg",      dtype = np.float64)
cast_unyt_array_mass_imperial     : Cast[unyt_array] = cast_unyt_array(unit = "lb",      dtype = np.float64)
cast_unyt_array_mass_cgs          : Cast[unyt_array] = cast_unyt_array(unit = "g",       dtype = np.float64)
cast_unyt_array_mass_solar        : Cast[unyt_array] = cast_unyt_array(unit = "Msun",    dtype = np.float64)
cast_unyt_array_time_milliseconds : Cast[unyt_array] = cast_unyt_array(unit = "ms",      dtype = np.float64)
cast_unyt_array_time_seconds      : Cast[unyt_array] = cast_unyt_array(unit = "s",       dtype = np.float64)
cast_unyt_array_time_minutes      : Cast[unyt_array] = cast_unyt_array(unit = "min",     dtype = np.float64)
cast_unyt_array_time_hours        : Cast[unyt_array] = cast_unyt_array(unit = "hr",      dtype = np.float64)
cast_unyt_array_time_days         : Cast[unyt_array] = cast_unyt_array(unit = "day",     dtype = np.float64)
cast_unyt_array_time_years        : Cast[unyt_array] = cast_unyt_array(unit = "yr",      dtype = np.float64)
cast_unyt_array_time_Gyrs         : Cast[unyt_array] = cast_unyt_array(unit = "Gyr",     dtype = np.float64)
cast_unyt_array_length_si         : Cast[unyt_array] = cast_unyt_array(unit = "m",       dtype = np.float64)
cast_unyt_array_length_imperial   : Cast[unyt_array] = cast_unyt_array(unit = "ft",      dtype = np.float64)
cast_unyt_array_length_cgs        : Cast[unyt_array] = cast_unyt_array(unit = "cm",      dtype = np.float64)
cast_unyt_array_length_solar      : Cast[unyt_array] = cast_unyt_array(unit = "Rsun",    dtype = np.float64)
cast_unyt_array_length_pc         : Cast[unyt_array] = cast_unyt_array(unit = "pc",      dtype = np.float64)
cast_unyt_array_length_kpc        : Cast[unyt_array] = cast_unyt_array(unit = "kpc",     dtype = np.float64)
cast_unyt_array_length_Mpc        : Cast[unyt_array] = cast_unyt_array(unit = "Mpc",     dtype = np.float64)
cast_unyt_array_speed_si          : Cast[unyt_array] = cast_unyt_array(unit = "m/s",     dtype = np.float64)
cast_unyt_array_speed_imperial    : Cast[unyt_array] = cast_unyt_array(unit = "ft/s",    dtype = np.float64)
cast_unyt_array_speed_cgs         : Cast[unyt_array] = cast_unyt_array(unit = "cm/s",    dtype = np.float64)
cast_unyt_array_speed_kms         : Cast[unyt_array] = cast_unyt_array(unit = "km/s",    dtype = np.float64)
cast_unyt_array_speed_Mpc_per_Gyr : Cast[unyt_array] = cast_unyt_array(unit = "Mpc/Gyr", dtype = np.float64)
cast_unyt_array_force_si          : Cast[unyt_array] = cast_unyt_array(unit = "N",       dtype = np.float64)
cast_unyt_array_force_imperial    : Cast[unyt_array] = cast_unyt_array(unit = "lbf",     dtype = np.float64)
cast_unyt_array_force_cgs         : Cast[unyt_array] = cast_unyt_array(unit = "dyn",     dtype = np.float64)
cast_unyt_array_energy_si         : Cast[unyt_array] = cast_unyt_array(unit = "J",       dtype = np.float64)
cast_unyt_array_energy_imperial   : Cast[unyt_array] = cast_unyt_array(unit = "BTU",     dtype = np.float64)
cast_unyt_array_energy_cgs        : Cast[unyt_array] = cast_unyt_array(unit = "erg",     dtype = np.float64)
cast_unyt_array_energy_eV         : Cast[unyt_array] = cast_unyt_array(unit = "eV",      dtype = np.float64)
cast_unyt_array_energy_KeV        : Cast[unyt_array] = cast_unyt_array(unit = "KeV",     dtype = np.float64)
cast_unyt_array_energy_GeV        : Cast[unyt_array] = cast_unyt_array(unit = "GeV",     dtype = np.float64)
cast_unyt_array_energy_TeV        : Cast[unyt_array] = cast_unyt_array(unit = "TeV",     dtype = np.float64)
cast_unyt_array_power_si          : Cast[unyt_array] = cast_unyt_array(unit = "W",       dtype = np.float64)
cast_unyt_array_power_imperial    : Cast[unyt_array] = cast_unyt_array(unit = "hp",      dtype = np.float64)
cast_unyt_array_power_cgs         : Cast[unyt_array] = cast_unyt_array(unit = "erg/s",   dtype = np.float64)
cast_unyt_array_power_solar       : Cast[unyt_array] = cast_unyt_array(unit = "Lsun",    dtype = np.float64)

# Nullables

cast_nullable_bool    : Cast[bool]    = nullable_cast(bool)
cast_nullable_int     : Cast[int]     = nullable_cast(int)
cast_nullable_float   : Cast[float]   = nullable_cast(float)
cast_nullable_Decimal : Cast[Decimal] = nullable_cast(Decimal)
cast_nullable_str     : Cast[str]     = nullable_cast(str)

cast_nullable_list  : Cast[List[Any]]  = nullable_cast(list)
cast_nullable_tuple : Cast[Tuple[Any, ...]] = nullable_cast(tuple)

def cast_nullable_typed_list(*type_casts: Cast[T]) -> NestedCast[List[T]]:
    return NestedCast[List[T]](cast_list, *type_casts, nullable = True)

def cast_nullable_typed_tuple(*type_casts: Cast[T]) -> NestedCast[Tuple[T]]:
    return NestedCast[Tuple[T]](cast_tuple, *type_casts, nullable = True)

def cast_nullable_dict(cast_keys: Union[Cast[K], None] = None, cast_values: Union[Cast[V], None] = None) -> Cast[Dict[K, V]]:
    if cast_keys is None:
        cast_keys = Cast.passthrough()
    if cast_values is None:
        cast_values = Cast.passthrough()
    def cast_nullable_dict_obj(value: Dict[Any, Any]) -> Dict[K, V]:
        return { cast(K, cast_keys(key)): cast(V, cast_values(value[key])) for key in value }
    return nullable_cast(cast_nullable_dict_obj)

def cast_nullable_ndarray(*args: Any, **kwargs: Any) -> Cast[np.ndarray]:
    """
    *args and **kwargs for numpy.array() may be passed
    """
    return nullable_cast(np.array, *args, **kwargs)

cast_nullable_ndarray_bool     : Cast[np.ndarray] = cast_nullable_ndarray(dtype = bool)
cast_nullable_ndarray_int      : Cast[np.ndarray] = cast_nullable_ndarray(dtype = int)
cast_nullable_ndarray_int8     : Cast[np.ndarray] = cast_nullable_ndarray(dtype = np.int8)
cast_nullable_ndarray_int16    : Cast[np.ndarray] = cast_nullable_ndarray(dtype = np.int16)
cast_nullable_ndarray_int32    : Cast[np.ndarray] = cast_nullable_ndarray(dtype = np.int32)
cast_nullable_ndarray_int64    : Cast[np.ndarray] = cast_nullable_ndarray(dtype = np.int64)
#cast_nullable_ndarray_int128   : Cast[np.ndarray] = cast_nullable_ndarray(dtype = np.int128)
#cast_nullable_ndarray_int256   : Cast[np.ndarray] = cast_nullable_ndarray(dtype = np.int256)
cast_nullable_ndarray_float    : Cast[np.ndarray] = cast_nullable_ndarray(dtype = float)
cast_nullable_ndarray_float16  : Cast[np.ndarray] = cast_nullable_ndarray(dtype = np.float16)
cast_nullable_ndarray_float32  : Cast[np.ndarray] = cast_nullable_ndarray(dtype = np.float32)
cast_nullable_ndarray_float64  : Cast[np.ndarray] = cast_nullable_ndarray(dtype = np.float64)
#cast_nullable_ndarray_float80  : Cast[np.ndarray] = cast_nullable_ndarray(dtype = np.float80)
#cast_nullable_ndarray_float96  : Cast[np.ndarray] = cast_nullable_ndarray(dtype = np.float96)
#cast_nullable_ndarray_float128 : Cast[np.ndarray] = cast_nullable_ndarray(dtype = np.float128)
#cast_nullable_ndarray_float256 : Cast[np.ndarray] = cast_nullable_ndarray(dtype = np.float256)
cast_nullable_ndarray_str      : Cast[np.ndarray] = cast_nullable_ndarray(dtype = str)

def cast_nullable_unyt_quantity(unit: str, *args: Any, **kwargs: Any) -> Cast[unyt_quantity]:
    """
    *args and **kwargs for unyt.unyt_quantity() may be passed
    (though will only be used if the target value is not already a unyt_quantity object!)
    """
    def inner(value: Any, _unit: str, *args: Any, **kwargs: Any):
        if isinstance(value, unyt_quantity):
            return value.to(_unit)
        elif isinstance(value, unyt_array):
            return unyt_quantity(value.to(_unit), _unit, *args, **kwargs)
        else:
            return unyt_quantity(value, _unit, *args, **kwargs)
    return nullable_cast(inner, unit, *args, **kwargs)

cast_nullable_unyt_quantity_mass_si           : Cast[unyt_quantity] = cast_nullable_unyt_quantity(unit = "kg",      dtype = np.float64)
cast_nullable_unyt_quantity_mass_imperial     : Cast[unyt_quantity] = cast_nullable_unyt_quantity(unit = "lb",      dtype = np.float64)
cast_nullable_unyt_quantity_mass_cgs          : Cast[unyt_quantity] = cast_nullable_unyt_quantity(unit = "g",       dtype = np.float64)
cast_nullable_unyt_quantity_mass_solar        : Cast[unyt_quantity] = cast_nullable_unyt_quantity(unit = "Msun",    dtype = np.float64)
cast_nullable_unyt_quantity_time_milliseconds : Cast[unyt_quantity] = cast_nullable_unyt_quantity(unit = "ms",      dtype = np.float64)
cast_nullable_unyt_quantity_time_seconds      : Cast[unyt_quantity] = cast_nullable_unyt_quantity(unit = "s",       dtype = np.float64)
cast_nullable_unyt_quantity_time_minutes      : Cast[unyt_quantity] = cast_nullable_unyt_quantity(unit = "min",     dtype = np.float64)
cast_nullable_unyt_quantity_time_hours        : Cast[unyt_quantity] = cast_nullable_unyt_quantity(unit = "hr",      dtype = np.float64)
cast_nullable_unyt_quantity_time_days         : Cast[unyt_quantity] = cast_nullable_unyt_quantity(unit = "day",     dtype = np.float64)
cast_nullable_unyt_quantity_time_years        : Cast[unyt_quantity] = cast_nullable_unyt_quantity(unit = "yr",      dtype = np.float64)
cast_nullable_unyt_quantity_time_Gyrs         : Cast[unyt_quantity] = cast_nullable_unyt_quantity(unit = "Gyr",     dtype = np.float64)
cast_nullable_unyt_quantity_length_si         : Cast[unyt_quantity] = cast_nullable_unyt_quantity(unit = "m",       dtype = np.float64)
cast_nullable_unyt_quantity_length_imperial   : Cast[unyt_quantity] = cast_nullable_unyt_quantity(unit = "ft",      dtype = np.float64)
cast_nullable_unyt_quantity_length_cgs        : Cast[unyt_quantity] = cast_nullable_unyt_quantity(unit = "cm",      dtype = np.float64)
cast_nullable_unyt_quantity_length_solar      : Cast[unyt_quantity] = cast_nullable_unyt_quantity(unit = "Rsun",    dtype = np.float64)
cast_nullable_unyt_quantity_length_pc         : Cast[unyt_quantity] = cast_nullable_unyt_quantity(unit = "pc",      dtype = np.float64)
cast_nullable_unyt_quantity_length_kpc        : Cast[unyt_quantity] = cast_nullable_unyt_quantity(unit = "kpc",     dtype = np.float64)
cast_nullable_unyt_quantity_length_Mpc        : Cast[unyt_quantity] = cast_nullable_unyt_quantity(unit = "Mpc",     dtype = np.float64)
cast_nullable_unyt_quantity_speed_si          : Cast[unyt_quantity] = cast_nullable_unyt_quantity(unit = "m/s",     dtype = np.float64)
cast_nullable_unyt_quantity_speed_imperial    : Cast[unyt_quantity] = cast_nullable_unyt_quantity(unit = "ft/s",    dtype = np.float64)
cast_nullable_unyt_quantity_speed_cgs         : Cast[unyt_quantity] = cast_nullable_unyt_quantity(unit = "cm/s",    dtype = np.float64)
cast_nullable_unyt_quantity_speed_kms         : Cast[unyt_quantity] = cast_nullable_unyt_quantity(unit = "km/s",    dtype = np.float64)
cast_nullable_unyt_quantity_speed_Mpc_per_Gyr : Cast[unyt_quantity] = cast_nullable_unyt_quantity(unit = "Mpc/Gyr", dtype = np.float64)
cast_nullable_unyt_quantity_force_si          : Cast[unyt_quantity] = cast_nullable_unyt_quantity(unit = "N",       dtype = np.float64)
cast_nullable_unyt_quantity_force_imperial    : Cast[unyt_quantity] = cast_nullable_unyt_quantity(unit = "lbf",     dtype = np.float64)
cast_nullable_unyt_quantity_force_cgs         : Cast[unyt_quantity] = cast_nullable_unyt_quantity(unit = "dyn",     dtype = np.float64)
cast_nullable_unyt_quantity_energy_si         : Cast[unyt_quantity] = cast_nullable_unyt_quantity(unit = "J",       dtype = np.float64)
cast_nullable_unyt_quantity_energy_imperial   : Cast[unyt_quantity] = cast_nullable_unyt_quantity(unit = "BTU",     dtype = np.float64)
cast_nullable_unyt_quantity_energy_cgs        : Cast[unyt_quantity] = cast_nullable_unyt_quantity(unit = "erg",     dtype = np.float64)
cast_nullable_unyt_quantity_energy_eV         : Cast[unyt_quantity] = cast_nullable_unyt_quantity(unit = "eV",      dtype = np.float64)
cast_nullable_unyt_quantity_energy_KeV        : Cast[unyt_quantity] = cast_nullable_unyt_quantity(unit = "KeV",     dtype = np.float64)
cast_nullable_unyt_quantity_energy_GeV        : Cast[unyt_quantity] = cast_nullable_unyt_quantity(unit = "GeV",     dtype = np.float64)
cast_nullable_unyt_quantity_energy_TeV        : Cast[unyt_quantity] = cast_nullable_unyt_quantity(unit = "TeV",     dtype = np.float64)
cast_nullable_unyt_quantity_power_si          : Cast[unyt_quantity] = cast_nullable_unyt_quantity(unit = "W",       dtype = np.float64)
cast_nullable_unyt_quantity_power_imperial    : Cast[unyt_quantity] = cast_nullable_unyt_quantity(unit = "hp",      dtype = np.float64)
cast_nullable_unyt_quantity_power_cgs         : Cast[unyt_quantity] = cast_nullable_unyt_quantity(unit = "erg/s",   dtype = np.float64)
cast_nullable_unyt_quantity_power_solar       : Cast[unyt_quantity] = cast_nullable_unyt_quantity(unit = "Lsun",    dtype = np.float64)

def cast_nullable_unyt_array(unit: str, *args: Any, **kwargs: Any) -> Cast[unyt_array]:
    """
    *args and **kwargs for unyt.unyt_array() may be passed
    (though will only be used if the target value is not already a unyt_array object!)
    """
    def inner(value: Any, _unit: str, *args: Any, **kwargs: Any):
        if isinstance(value, unyt_quantity):
            return unyt_array(value.to(_unit), _unit, *args, **kwargs)
        elif isinstance(value, unyt_array):
            return value.to(_unit)
        else:
            return unyt_array(value, _unit, *args, **kwargs)
    return nullable_cast(inner, unit, *args, **kwargs)

cast_nullable_unyt_array_mass_si           : Cast[unyt_array] = cast_nullable_unyt_array(unit = "kg",      dtype = np.float64)
cast_nullable_unyt_array_mass_imperial     : Cast[unyt_array] = cast_nullable_unyt_array(unit = "lb",      dtype = np.float64)
cast_nullable_unyt_array_mass_cgs          : Cast[unyt_array] = cast_nullable_unyt_array(unit = "g",       dtype = np.float64)
cast_nullable_unyt_array_mass_solar        : Cast[unyt_array] = cast_nullable_unyt_array(unit = "Msun",    dtype = np.float64)
cast_nullable_unyt_array_time_milliseconds : Cast[unyt_array] = cast_nullable_unyt_array(unit = "ms",      dtype = np.float64)
cast_nullable_unyt_array_time_seconds      : Cast[unyt_array] = cast_nullable_unyt_array(unit = "s",       dtype = np.float64)
cast_nullable_unyt_array_time_minutes      : Cast[unyt_array] = cast_nullable_unyt_array(unit = "min",     dtype = np.float64)
cast_nullable_unyt_array_time_hours        : Cast[unyt_array] = cast_nullable_unyt_array(unit = "hr",      dtype = np.float64)
cast_nullable_unyt_array_time_days         : Cast[unyt_array] = cast_nullable_unyt_array(unit = "day",     dtype = np.float64)
cast_nullable_unyt_array_time_years        : Cast[unyt_array] = cast_nullable_unyt_array(unit = "yr",      dtype = np.float64)
cast_nullable_unyt_array_time_Gyrs         : Cast[unyt_array] = cast_nullable_unyt_array(unit = "Gyr",     dtype = np.float64)
cast_nullable_unyt_array_length_si         : Cast[unyt_array] = cast_nullable_unyt_array(unit = "m",       dtype = np.float64)
cast_nullable_unyt_array_length_imperial   : Cast[unyt_array] = cast_nullable_unyt_array(unit = "ft",      dtype = np.float64)
cast_nullable_unyt_array_length_cgs        : Cast[unyt_array] = cast_nullable_unyt_array(unit = "cm",      dtype = np.float64)
cast_nullable_unyt_array_length_solar      : Cast[unyt_array] = cast_nullable_unyt_array(unit = "Rsun",    dtype = np.float64)
cast_nullable_unyt_array_length_pc         : Cast[unyt_array] = cast_nullable_unyt_array(unit = "pc",      dtype = np.float64)
cast_nullable_unyt_array_length_kpc        : Cast[unyt_array] = cast_nullable_unyt_array(unit = "kpc",     dtype = np.float64)
cast_nullable_unyt_array_length_Mpc        : Cast[unyt_array] = cast_nullable_unyt_array(unit = "Mpc",     dtype = np.float64)
cast_nullable_unyt_array_speed_si          : Cast[unyt_array] = cast_nullable_unyt_array(unit = "m/s",     dtype = np.float64)
cast_nullable_unyt_array_speed_imperial    : Cast[unyt_array] = cast_nullable_unyt_array(unit = "ft/s",    dtype = np.float64)
cast_nullable_unyt_array_speed_cgs         : Cast[unyt_array] = cast_nullable_unyt_array(unit = "cm/s",    dtype = np.float64)
cast_nullable_unyt_array_speed_kms         : Cast[unyt_array] = cast_nullable_unyt_array(unit = "km/s",    dtype = np.float64)
cast_nullable_unyt_array_speed_Mpc_per_Gyr : Cast[unyt_array] = cast_nullable_unyt_array(unit = "Mpc/Gyr", dtype = np.float64)
cast_nullable_unyt_array_force_si          : Cast[unyt_array] = cast_nullable_unyt_array(unit = "N",       dtype = np.float64)
cast_nullable_unyt_array_force_imperial    : Cast[unyt_array] = cast_nullable_unyt_array(unit = "lbf",     dtype = np.float64)
cast_nullable_unyt_array_force_cgs         : Cast[unyt_array] = cast_nullable_unyt_array(unit = "dyn",     dtype = np.float64)
cast_nullable_unyt_array_energy_si         : Cast[unyt_array] = cast_nullable_unyt_array(unit = "J",       dtype = np.float64)
cast_nullable_unyt_array_energy_imperial   : Cast[unyt_array] = cast_nullable_unyt_array(unit = "BTU",     dtype = np.float64)
cast_nullable_unyt_array_energy_cgs        : Cast[unyt_array] = cast_nullable_unyt_array(unit = "erg",     dtype = np.float64)
cast_nullable_unyt_array_energy_eV         : Cast[unyt_array] = cast_nullable_unyt_array(unit = "eV",      dtype = np.float64)
cast_nullable_unyt_array_energy_KeV        : Cast[unyt_array] = cast_nullable_unyt_array(unit = "KeV",     dtype = np.float64)
cast_nullable_unyt_array_energy_GeV        : Cast[unyt_array] = cast_nullable_unyt_array(unit = "GeV",     dtype = np.float64)
cast_nullable_unyt_array_energy_TeV        : Cast[unyt_array] = cast_nullable_unyt_array(unit = "TeV",     dtype = np.float64)
cast_nullable_unyt_array_power_si          : Cast[unyt_array] = cast_nullable_unyt_array(unit = "W",       dtype = np.float64)
cast_nullable_unyt_array_power_imperial    : Cast[unyt_array] = cast_nullable_unyt_array(unit = "hp",      dtype = np.float64)
cast_nullable_unyt_array_power_cgs         : Cast[unyt_array] = cast_nullable_unyt_array(unit = "erg/s",   dtype = np.float64)
cast_nullable_unyt_array_power_solar       : Cast[unyt_array] = cast_nullable_unyt_array(unit = "Lsun",    dtype = np.float64)



from .Validators import isURI, isURL, isURIScheme, isURIAuthority, isURIPath, isURIQuery, isURIFragment

class URI(object):
    """
    A Universal Resourse Indicator

    For detailed infomation visit https://en.wikipedia.org/wiki/Uniform_Resource_Identifier
    """

    class Scheme(object):
        def __init__(self, value):
            self.value = value

        def isValid(self):
            return isURIScheme(str(self))

        def isSecure(self):
            return self.value == "https:"
        
        def __str__(self):
            return self.value

    class Authority(object):
        def __init__(self, value):
            atSplit = value.split("@")
            self.userinfo = atSplit[0] if len(atSplit) > 1 else ""

            colonSplit = value.split(":")
            self.host = colonSplit[1 if len(atSplit) > 1 else 0]

            self.port = int(colonSplit[2 if len(atSplit) > 1 else 1]) if len(colonSplit) > 2 or (len(colonSplit) > 1 and len(atSplit) == 1) else None

        def isValid(self):
            return isURIAuthority(str(self))

        def __str__(self):
            return (self.userinfo + "@") if self.userinfo != "" else "" + self.host + (":" + str(self.port)) if self.port is not None else ""

    class Path(object):
        def __init__(self, value):
            self.value = value

        def isValid(self):
            return isURIPath(str(self))

        def __str__(self):
            return self.value

    class Query(object):
        def __init__(self, value):
            self.value = value

        def isValid(self):
            return isURIQuery(str(self))

        def __str__(self):
            return self.value

    class Fragment(object):
        def __init__(self, value):
            self.value = value

        def isValid(self):
            return isURIFragment(str(self))

        def __str__(self):
            return self.value

    def __init__(self, scheme, authority = URI.Authority(""), path = URI.Path("/"), query = URI.Query(""), fragment = URI.Fragment("")):
        self.scheme = scheme if isinstance(scheme, URI.Scheme) else URI.Scheme(scheme)
        self.authority = authority if isinstance(authority, URI.Authority) else URI.Authority(authority)
        self.path = path if isinstance(path, URI.Path) else URI.Path(path)
        self.query = query if isinstance(query, URI.Query) else URI.Query(query)
        self.fragment = fragment if isinstance(fragment, URI.Fragment) else URI.Fragment(fragment)

    def isValidURI(self):
        return isURI(str(self))

    def isValidURL(self):
        return isURL(str(self))

    def isValid(self):
        return self.isValidURI()

    def __str__(self):
        return str(self.scheme) + str(self.authority) + str(self.path) + str(self.query) + str(self.fragment)

    @staticmethod
    def fromString(string):
        pass

class URL(URI):
    def __init__(self, secure = True, authority = URI.Authority(""), path = URI.Path("/"), query = URI.Query(""), fragment = URI.Fragment("")):
        super().__init__("https:" if secure else "http:", authority, path, query, fragment)

    def isValid(self):
        return self.isValidURL()


import warnings

from .._global_settings import settings_object as __settings_object



# Try to load the mpi4py package and catch any errors
try:
    from mpi4py import MPI
    __settings_object._set_mpi_avalible()
    _display_mpi_warning = lambda *args, **kwargs: None
    _raise_mpi_error = lambda *args, **kwargs: None
except:
    def _display_mpi_warning(*args, **kwargs):
        warnings.warn("Either mpi4py is not avalible, the program was not started for use with MPI or mpi4py failed to install/load correctly. Disabling optional MPI functionality. Non-optional MPI functionality is not avalible.", ImportWarning)
    def _raise_mpi_error(*args, **kwargs):
        raise ImportError("Either mpi4py is not avalible, the program was not started for use with MPI or mpi4py failed to install/load correctly.")



# Load the stuff that dosen't depend on the rest of the package
if __settings_object.mpi_avalible:
    from ._independant_mpi._mpi_config import mpi_config as MPI_Config
    def get_mpi_rank() -> int:
        return MPI_Config.rank
else:
    from ._independant_dummy._mpi_config import mpi_config as MPI_Config
    # This dummy function needs to be here to avoid circular imports when all that is needed is the current rank
    def get_mpi_rank() -> int:
        return 0



# Now import stuff that has dependancies from elsewhere in tha package
if __settings_object.mpi_avalible:
    from ._dependant_mpi import *
else:
    # Load everything else from the _dummy subpackage instead
    from ._dependant_dummy import *



from typing import Any, TypeVar, ParamSpec
from collections.abc import Callable, Sized, Sequence
from functools import wraps

import numpy as np

from ...Tools._edit_locals import use_locals



P = ParamSpec("P")
T = TypeVar("T")



@use_locals(0, 0)
def synchronyse(target: T, /, root: int = 0, comm: object|None = None) -> T:
    """
    Broadcast the value of a variable on the root rank to the same variable on all ranks in the communicator.

    The first argument should be a string containing the variable name NOT the value of the variable!
    """
    return target



def mpi_check_equal(value: T, /, root: int|None = None, comm: object|None = None) -> bool:
    """
    Check all ranks have the same value.
    """
    return True



def if_mpi_root(default: T, root: int|None = None):
    """
    Function decorator for restricting calls to the root rank.

    A default return value of the appropriate return type must be provided (this can be None if the wrapped function allows it or specifies no return type).
    """
    def inner(func: Callable[P, T]) -> Callable[P, T]:
        return func
    return inner



def mpi_barrier(comm: object|None = None) -> None:
    """
    Wait for all ranks in the communicator to reach this barrier.
    """
    pass



def mpi_sum(data: Sequence[T], comm: object|None = None, root: int|None = None) -> T:
    """
    Calculate the sum of data across multiple ranks.
    """
    if len(data) > 0:
        raise IndexError("No data provided by any rank.")
    return sum(data[1:], start = data[0]) if len(data) > 1 else data[0]



def mpi_mean(data: Sequence[float], weights: Sequence[float]|None = None, comm: object|None = None, root: int|None = None) -> float:
    """
    Calculate the mean of data across multiple ranks.
    Optionally, specify weights to calculate the weighted mean.
    """
    return float(np.average(np.array(data), weights = np.array(weights)))



def mpi_get_slice(length: int|Sized, comm: object|None = None, rank: int|None = None) -> slice:
    """
    Partition a set of data such that it is distributed evenly accross each MPI rank.

    This method can be used to help get data from a h5py dataset when using parallel IO.

    Parameters:
                 int|Sized length -> The length of the target data or the target data if a size can be derived from it.
        MPI.Intracomm|None comm   -> Optional MPI communicator object (defaults to the one from MPI_Config).
                  int|None comm   -> Optional MPI rank (defaults to the one from MPI_Config if not comm is specified or the one from the provided comm).

    Returns:
        slice -> A slice object that can be applied a Sequence of the appropriate length.
    """
    if not isinstance(length, int):
        length = len(length)
    return slice(0, length)



def mpi_slice(data: Sequence[T], comm: object|None = None, rank: int|None = None) -> Sequence[T]:
    """
    Partition a set of data such that it is distributed evenly accross each MPI rank.

    This method can be used to quickly get data from a h5py dataset when using parallel IO.

    Parameters:
                  Sequence data -> Data to be partitioned.
        MPI.Intracomm|None comm -> Optional MPI communicator object (defaults to the one from MPI_Config).
                  int|None comm -> Optional MPI rank (defaults to the one from MPI_Config if not comm is specified or the one from the provided comm).

    Returns:
        Sequence -> Section of the data assigned to the calling rank.
    """
    return data[:]



def mpi_gather_array(data: np.ndarray, comm: object|None = None, root: int|None = None, target_buffer: np.ndarray|None = None, allgather: bool = False) -> np.ndarray|None|tuple[np.ndarray, np.ndarray]|tuple[None, None]:
    """
    Gather numpy array data to the root rank.
    If the resulting array is larger than the MPI buffer length, data will be transmitted point-to-point in rank order instead of using the Gatherv method.
    """
    if target_buffer is None:
        return data
    else:
        target_buffer[:] = data[:]
        return target_buffer



def mpi_scatter_array(data: np.ndarray|None, elements_this_rank: int|None = None, elements_per_rank: list[int]|None = None, comm: object|None = None, root: int|None = None, target_buffer: np.ndarray|None = None) -> np.ndarray:
    """
    Scatter numpy array data from the root rank to all ranks.
    If the target array is larger than the MPI buffer length, data will be transmitted point-to-point in rank order instead of using the Scatterv method.
    """
    if elements_this_rank is not None and elements_this_rank != data.shape[0]:
        raise

    if target_buffer is None:
        return data
    else:
        target_buffer[:] = data[:]
        return target_buffer



def mpi_redistribute_array_evenly(data: np.ndarray, elements_this_rank: int|None = None, elements_per_rank: list[int]|None = None, comm: object|None = None, root: int|None = None, target_buffer: np.ndarray|None = None) -> np.ndarray:
    return data


def mpi_argsort(
    data: np.ndarray,
    use_index_dtype = np.int64,
    output_destination_ranks: bool = False,
    output_inplace_target_indexes: bool = False
) -> np.ndarray|tuple[np.ndarray, np.ndarray]|tuple[np.ndarray, np.ndarray, np.ndarray]:
    sorted_indexes = np.argsort(data)
    result = [sorted_indexes]
    if output_destination_ranks:
        result.append(np.zeros(shape = data.shape, dtype = np.int64))
    if output_inplace_target_indexes:
        result.append(sorted_indexes)
    return result[0] if len(result) == 1 else tuple(result)



def mpi_sort(
        data: np.ndarray,
        order: tuple[np.ndarray, np.ndarray]|None = None,
        result_data_buffer: np.ndarray|None = None,
        assume_memory_limitation: bool = False
) -> np.ndarray:
    """
    Sort a set of data in parallel.

    Parameters:

                           `numpy.ndarray` `data`                     -> The data to be sorted (along axis = 0)

        (`numpy.ndarray`, `numpy.ndarray`) `order`                    -> The ordering information from `mpi_argsort(...)[1:]` (target_ranks, inplace_sort_indexes) (optional)

                           `numpy.ndarray` `result_data_buffer`       -> Buffer to place results into (optional)

                                    `bool` `assume_memory_limitation` -> Perform the sort in a memory-efficient way, at the cost of speed (defaults to False - fastest)

    Returns:
        `numpy.ndarray` -> The data in a sorted order distributed accross each rank using the same distribution as the input data
    """
    result = np.sort(data)
    if result_data_buffer is not None:
        result_data_buffer[...] = result[...]
    return result



def mpi_calculate_reorder(
    input_ids: np.ndarray,
    output_ids: np.ndarray,
    assume_memory_limitation: bool = False
) -> tuple[np.ndarray, np.ndarray]:
    order = mpi_argsort(input_ids, output_destination_ranks = True, output_inplace_target_indexes = True)
    reverse_order = mpi_argsort(order[0], output_destination_ranks = True, output_inplace_target_indexes = True)
    del order

    target_sort_order = mpi_argsort(output_ids, output_destination_ranks = True, output_inplace_target_indexes = True)
    reverse_target_sort_order = mpi_argsort(target_sort_order[0], output_destination_ranks = True, output_inplace_target_indexes = True)
    del target_sort_order

    return (
        mpi_sort(data = reverse_target_sort_order[1], order = reverse_order[1:], assume_memory_limitation = assume_memory_limitation),
        mpi_sort(data = reverse_target_sort_order[2], order = reverse_order[1:], assume_memory_limitation = assume_memory_limitation)
    )



def mpi_reorder(
    input_ids: np.ndarray,
    input_data: np.ndarray,
    output_ids: np.ndarray,
    output_data_buffer: np.ndarray|None = None,
    assume_memory_limitation: bool = False
) -> np.ndarray:
    order = mpi_calculate_reorder(input_ids, output_ids)
    return mpi_sort(
        data = input_data,
        order = order,
        result_data_buffer = output_data_buffer,
        assume_memory_limitation = assume_memory_limitation
    )



def mpi_apply_reorder(
    input_data: np.ndarray,
    order: tuple[np.ndarray, np.ndarray],
    output_data_buffer: np.ndarray|None = None,
    assume_memory_limitation: bool = False
) -> np.ndarray:
    """
    Wrapper for `mpi_sort` using a subset of the parameter names from `mpi_reorder`.

    `input_data` -> `data`

    `order` -> `order`

    `output_data_buffer` -> `result_data_buffer`

    `assume_memory_limitation` -> `assume_memory_limitation`
    """
    return mpi_sort(
        data = input_data,
        order = order,
        result_data_buffer = output_data_buffer,
        assume_memory_limitation = assume_memory_limitation
    )




from typing import TYPE_CHECKING, Literal

from matplotlib.axes import Axes
if TYPE_CHECKING:
    from matplotlib.pylab import ArrayLike
from matplotlib.typing import ColorType
import numpy as np
from matplotlib import pyplot as plt
from matplotlib.contour import QuadContourSet

from ..Data._Rect import Rect
from ._CachedPlotElements import CachedPlotContour

class Contour(object):

    def __init__(self, x_data: np.ndarray[tuple[int], np.dtype[np.floating]], y_data: np.ndarray[tuple[int], np.dtype[np.floating]], values: np.ndarray[tuple[int], np.dtype[np.floating]]|None = None) -> None:
        self.__can_compute: bool = True
        self.__x_data = x_data
        self.__y_data = y_data
        self.__values = values

        #self.__contour_levels: set[float] = set()
        self.__contour_levels: dict[float, list[ColorType, ArrayLike, Literal["solid", "dashed", "dashdot", "dotted"], float]] = {}

        self.__can_plot: bool = False
        self.__x_bin_locations: np.ndarray[tuple[int], np.dtype[np.floating]]|None = None
        self.__y_bin_locations: np.ndarray[tuple[int], np.dtype[np.floating]]|None = None
        self.__levels: np.ndarray[tuple[int, int], np.dtype[np.floating]]|None = None
        self.__cache_object = CachedPlotContour()
        self.__result: QuadContourSet|None = None

    def release_resources(self) -> None:
        if self.__can_compute:
            self.__can_compute = False
            del self.__x_data
            del self.__y_data
            del self.__values

    @property
    def data(self) -> CachedPlotContour:
        """
        Get the cache object for this contour.
        """
        return self.__cache_object

    @property
    def plotting_result(self) -> QuadContourSet|None:
        """
        Get the plotting result for this contour.
        """
        return self.__result

    @property
    def levels(self) -> np.ndarray[tuple[int, int], np.dtype[np.floating]]|None:
        return self.__levels

    @property
    def min_level(self) -> float|None:
        return self.__levels.min() if self.__levels is not None else None

    @property
    def max_level(self) -> float|None:
        return self.__levels.max() if self.__levels is not None else None

    @property
    def contour_levels(self) -> set[float]:
        return set(self.__contour_levels.keys())

    @property
    def contour_colours(self) -> tuple[ColorType, ...]:
        return tuple(self.__contour_levels[level][0] for level in sorted(self.__contour_levels.keys()))
    
    @property
    def contour_linewidths(self) -> tuple["ArrayLike", ...]:
        return tuple(self.__contour_levels[level][1] for level in sorted(self.__contour_levels.keys()))
    
    @property
    def contour_linestyles(self) -> tuple[Literal["solid", "dashed", "dashdot", "dotted"], ...]:
        return tuple(self.__contour_levels[level][2] for level in sorted(self.__contour_levels.keys()))
    
    @property
    def contour_alphas(self) -> tuple[float, ...]:
        return tuple(self.__contour_levels[level][3] for level in sorted(self.__contour_levels.keys()))
    
    def add_contours(self, *levels: float) -> None:
        for level in levels:
            if level not in self.__contour_levels:
                self.__contour_levels[level] = ["black", 1, "solid", 1.0]

    def remove_contours(self, *levels: float) -> None:
        for level in levels:
            if level in self.__contour_levels:
                del self.__contour_levels[level]
    
    def clear_contours(self) -> None:
        self.__contour_levels.clear()

    def set_colour(self, level: float, colour: ColorType) -> None:
        if level in self.__contour_levels:
            self.__contour_levels[level][0] = colour
        else:
            raise ValueError(f"Contour level {level} not found.")
        
    def set_linewidth(self, level: float, linewidth: "ArrayLike") -> None:
        if level in self.__contour_levels:
            self.__contour_levels[level][1] = linewidth
        else:
            raise ValueError(f"Contour level {level} not found.")
        
    def set_linestyle(self, level: float, linestyle: Literal["solid", "dashed", "dashdot", "dotted"]) -> None:
        if level in self.__contour_levels:
            self.__contour_levels[level][2] = linestyle
        else:
            raise ValueError(f"Contour level {level} not found.")

    def set_alpha(self, level: float, alpha: float) -> None:
        if level in self.__contour_levels:
            self.__contour_levels[level][3] = alpha
        else:
            raise ValueError(f"Contour level {level} not found.")

    def generate(self, gridsize = int|tuple[int,int], extent: Rect|None = None, **kwargs) -> None:
        if not self.__can_compute:
            raise RuntimeError("Cannot compute contours after resources have been released.")

        hist, bin_x, bin_y = np.histogram2d(
            self.__x_data,
            self.__y_data,
            weights = self.__values,
            bins = gridsize,
            range = extent.range if extent is not None else None
        )

        self.__levels = hist.T

        self.__x_bin_locations = bin_x[:-1] + (bin_x[1] - bin_x[0]) / 2
        self.__y_bin_locations = bin_y[:-1] + (bin_y[1] - bin_y[0]) / 2

        self.__can_plot = True

        self.__cache_object = CachedPlotContour()
        self.__cache_object.x = self.__x_bin_locations
        self.__cache_object.y = self.__y_bin_locations
        self.__cache_object.z = self.__levels
        self.__cache_object.levels = tuple(sorted(self.__contour_levels))
        self.__cache_object.linewidths = self.contour_linewidths
        self.__cache_object.linestyles = self.contour_linestyles
        self.__cache_object.alpha_values = self.contour_alphas
        self.__cache_object.colours = self.contour_colours

    def plot(self, axis: Axes, **kwargs) -> QuadContourSet:
        if not self.__can_plot:
            raise RuntimeError("Cannot plot contours before calling \"generate\".")

        self.__result = axis.contour(
            self.__x_bin_locations,
            self.__y_bin_locations,
            self.__levels,
            levels = sorted(self.__contour_levels),
            linewidths = self.contour_linewidths,
            linestyles = self.contour_linestyles,
            alpha = self.contour_alphas,
            colors = self.contour_colours,
            **kwargs
        )

        #self.__cache_object = CachedPlotContour.from_contours(self.__result, self.__x_bin_locations, self.__y_bin_locations, self.__levels)
        self.__cache_object._result = self.__result

        return self.__result



from typing import TYPE_CHECKING, Any, Literal, Sequence, TypeVar, Generic
from abc import abstractmethod

from matplotlib.axes import Axes
from matplotlib.cm import ScalarMappable
from matplotlib.colorbar import Colorbar
from matplotlib.colorizer import ColorizingArtist
from matplotlib.colors import Colormap
from matplotlib.figure import Figure
if TYPE_CHECKING:
    from matplotlib.pylab import ArrayLike
from matplotlib.typing import ColorType, MarkerType
import numpy as np
from matplotlib.collections import PathCollection, PolyCollection
from matplotlib.container import ErrorbarContainer
from matplotlib.lines import Line2D
from matplotlib.contour import QuadContourSet

from ..Data._Rect import Rect
from ..Tools._Struct import CacheableStruct
from ..Tools._autoproperty import AutoProperty, AutoProperty_NonNullable

T = TypeVar("T")

class CachedPlotElement(CacheableStruct, Generic[T]):
    _result = AutoProperty[T](allow_uninitialised = True)
    def __init__(self, *cacheable_attributes: str, **kwargs):
        super().__init__(
            cacheable_attributes = cacheable_attributes,#list(cacheable_attributes) + ["_result"],
            **kwargs
        )
    @property
    def result(self) -> T|None:
        return self._result
    @abstractmethod
    def render(self, figure: Figure, axis: Axes, *args: Any, **kwargs: Any) -> None:
        """
        Render the element on the given figure and axis.
        
        Parameters:
            matplotlib.figure.Figure figure:
                The figure to render on.
            matplotlib.axes.Axes axis:
                The axis to render on.
        """
        raise NotImplementedError("Subclasses must implement this method.")

class CachedPlotLine(CachedPlotElement[Line2D]):
    x = AutoProperty_NonNullable[np.ndarray[tuple[int], np.dtype[np.floating]]]()
    y = AutoProperty_NonNullable[np.ndarray[tuple[int], np.dtype[np.floating]]]()
    label = AutoProperty[str](allow_uninitialised = True)
    colour = AutoProperty[ColorType](allow_uninitialised = True)
    linestyle = AutoProperty[str](allow_uninitialised = True)
    linewidth = AutoProperty[float](allow_uninitialised = True)
    alpha = AutoProperty_NonNullable[float](default_value = 1.0)
    def __init__(self, **kwargs):
        super().__init__("x", "y", "label", "colour", "linestyle", "linewidth", "alpha", **kwargs)
    def render(self, figure: Figure, axis: Axes, *args: Any, **kwargs: Any) -> None:
        self._result = axis.plot(
            self.x,
            self.y,
            label = self.label,
            color = self.colour,
            linestyle = self.linestyle,
            linewidth = self.linewidth,
            alpha = self.alpha,
            **kwargs
        )[0]
    @staticmethod
    def from_line(line: Line2D) -> "CachedPlotLine":
        coords: np.ndarray[tuple[int, int], np.dtype[np.floating]] = line.get_xydata()
        return CachedPlotLine(
            x = coords[:, 0],
            y = coords[:, 1],
            label = line.get_label(),
            colour = line.get_color(),
            linestyle = line.get_linestyle(),
            linewidth = line.get_linewidth(),
            alpha = line.get_alpha(),
            _result = line
        )

class CachedPlotScatter(CachedPlotElement[PathCollection]):
    x = AutoProperty_NonNullable[np.ndarray[tuple[int], np.dtype[np.floating]]]()
    y = AutoProperty_NonNullable[np.ndarray[tuple[int], np.dtype[np.floating]]]()
    label = AutoProperty[str](allow_uninitialised = True)
    colour = AutoProperty[ColorType|Sequence[ColorType]](allow_uninitialised = True)
    colourmap = AutoProperty[str|Colormap](allow_uninitialised = True)
    marker = AutoProperty[MarkerType](allow_uninitialised = True)
    size = AutoProperty["float|ArrayLike"](allow_uninitialised = True)
    alpha = AutoProperty["float|ArrayLike"](allow_uninitialised = True)
    def __init__(self, **kwargs):
        super().__init__("x", "y", "label", "colour", "colourmap", "marker", "size", "alpha", **kwargs)
    def render(self, figure: Figure, axis: Axes, *args: Any, **kwargs: Any):
        self._result = axis.scatter(self.x, self.y, c = self.colour, cmap = self.colourmap, marker = self.marker, s = self.size, alpha = self.alpha, label = self.label, **kwargs)
    @staticmethod
    def from_points(points: PathCollection) -> "CachedPlotScatter":
        coords: np.ndarray[tuple[int, int], np.dtype[np.floating]] = points.get_offsets()
        return CachedPlotScatter(
            x = coords[:, 0],
            y = coords[:, 1],
            label = points.get_label(),
            colour = points.get_facecolor(),
            colourmap = points.get_cmap(),
            marker = points.get_paths()[0],
            size = points.get_sizes(),
            alpha = points.get_alpha(),
            _result = points
        )

class CachedPlotErrorbar(CachedPlotElement[ErrorbarContainer]):
    x = AutoProperty_NonNullable[np.ndarray[tuple[int], np.dtype[np.floating]]]()
    y = AutoProperty_NonNullable[np.ndarray[tuple[int], np.dtype[np.floating]]]()
    xerr = AutoProperty[np.ndarray[tuple[int], np.dtype[np.floating]]](allow_uninitialised = True)
    yerr = AutoProperty[np.ndarray[tuple[int], np.dtype[np.floating]]](allow_uninitialised = True)
    label = AutoProperty[str](allow_uninitialised = True)
    line_colour = AutoProperty[ColorType|Sequence[ColorType]](allow_uninitialised = True)
    marker_colour = AutoProperty[ColorType|Sequence[ColorType]](allow_uninitialised = True)
    error_colour = AutoProperty[ColorType|Sequence[ColorType]](allow_uninitialised = True)
    error_line_width = AutoProperty[float](allow_uninitialised = True)
    colourmap = AutoProperty[str|Colormap](allow_uninitialised = True)
    marker = AutoProperty[MarkerType](allow_uninitialised = True, default_value = "")
    size = AutoProperty["float|ArrayLike"](allow_uninitialised = True)
    alpha = AutoProperty["float|ArrayLike"](allow_uninitialised = True)
    def __init__(self, **kwargs):
        super().__init__("x", "y", "xerr", "yerr", "colourmap", **kwargs)
    def render(self, figure: Figure, axis: Axes, *args: Any, **kwargs: Any):
        self._result = axis.errorbar(
            x = self.x,
            y = self.y,
            yerr = self.yerr,
            xerr = self.xerr,
            label = self.label,
            ecolor = self.error_colour,
            elinewidth = self.error_line_width,
            cmap = self.colourmap,
            marker = self.marker,
            markersize = self.size,
            alpha = self.alpha,
            color = self.line_colour,
            **kwargs
        )
    @staticmethod
    def from_errorbar(bars: ErrorbarContainer) -> "CachedPlotErrorbar":
        coords: np.ndarray[tuple[int, int], np.dtype[np.floating]] = bars.get_offsets()
        return CachedPlotErrorbar(
            x = coords[:, 0],
            y = coords[:, 1],
            label = bars.get_label(),
            colour = bars.get_facecolor(),
            colourmap = bars.get_cmap(),
            marker = bars.get_paths()[0],
            size = bars.get_sizes(),
            alpha = bars.get_alpha(),
            _result = bars
        )

class CachedPlotHexbin(CachedPlotElement[PolyCollection]):
    extent = AutoProperty_NonNullable[Rect]()
    gridsize = AutoProperty_NonNullable[int|tuple[int, int]]()
    polygon_offsets = AutoProperty_NonNullable[np.ndarray[tuple[int, int], np.dtype[np.floating]]]()
    bin_values = AutoProperty_NonNullable[np.ndarray[tuple[int], np.dtype[np.floating]]]()
    min_value = AutoProperty[float](allow_uninitialised = True)
    max_value = AutoProperty[float](allow_uninitialised = True)
    colourmap = AutoProperty[str|Colormap](allow_uninitialised = True)
    edgecolour = AutoProperty_NonNullable[ColorType|Literal["face", "none"]](default_value = "face")
    bin_alphas = AutoProperty[np.ndarray[tuple[int], np.dtype[np.floating]]](allow_uninitialised = True)
    def __init__(self, **kwargs):
        super().__init__("extent", "gridsize", "polygon_offsets", "bin_values", "min_value", "max_value", "colourmap", "edgecolour", "bin_alphas", **kwargs)
    def render(self, figure: Figure, axis: Axes, *args: Any, **kwargs: Any):
        self._result = axis.hexbin(
            x = self.polygon_offsets[:, 0],
            y = self.polygon_offsets[:, 1],
            C = self.bin_values,
            gridsize = self.gridsize,
            cmap = self.colourmap,
            vmin = self.min_value,
            vmax = self.max_value,
            edgecolor = self.edgecolour,
            extent = self.extent.extent,
            **kwargs
        )
        self._result.set_alpha(self.bin_alphas) # type: ignore[arg-type]
    @staticmethod
    def from_hexes(hexes: PolyCollection, extent: Rect, gridsize: int|tuple[int, int], min_value: float|None = None, max_value: float|None = None, edgecolours_match_faces: bool = False) -> "CachedPlotHexbin":
        n_hexes = hexes.get_array().shape[0]
        edgecolour = hexes.get_edgecolor()
        if edgecolours_match_faces:
            edgecolour = "face"
        elif isinstance(edgecolour, np.ndarray) and edgecolour.shape[0] == 0 and n_hexes > 0:
            edgecolour = "none"
        return CachedPlotHexbin(
            extent = extent,
            gridsize = gridsize,
            polygon_offsets = hexes.get_offsets(),
            bin_values = hexes.get_array(),
            min_value = min_value,
            max_value = max_value,
            edgecolour = edgecolour,
            colourmap = hexes.get_cmap(),
            bin_alphas = hexes.get_alpha(),
            _result = hexes
        )

class CachedPlotColourbar(CachedPlotElement[Colorbar]):
    target_element = AutoProperty_NonNullable[str]()
    label =  AutoProperty[str](allow_uninitialised = True)
    location =  AutoProperty[Literal["left", "right", "top", "bottom"]](default_value = "right")
    extend = AutoProperty[Literal["neither", "both", "min", "max"]](default_value = "neither")
    def __init__(self, **kwargs):
        super().__init__("target_element", "label", "location", "extend", **kwargs)
    def render(self, figure: Figure, axis: Axes, target: ScalarMappable|ColorizingArtist, *args: Any, **kwargs: Any) -> None:
        """
        Render the element on the given figure and axis, using the target element.
        
        Parameters:
            matplotlib.figure.Figure figure:
                The figure to render on.
            matplotlib.axes.Axes axis:
                The axis to render on.
            ScalarMappable|ColorizingArtist target:
                The target element for the colourbar.
        """
        self._result = figure.colorbar(
            target,
            ax = axis,
            location = self.location,
            label = self.label,
            extend = self.extend
        )

class CachedPlotContour(CachedPlotElement[QuadContourSet]):
    x = AutoProperty_NonNullable[np.ndarray[tuple[int], np.dtype[np.floating]]]()
    y = AutoProperty_NonNullable[np.ndarray[tuple[int], np.dtype[np.floating]]]()
    z = AutoProperty_NonNullable[np.ndarray[tuple[int, int], np.dtype[np.floating]]]()
    levels = AutoProperty_NonNullable[tuple[float]]()
    linewidths = AutoProperty["ArrayLike"](allow_uninitialised = True)
    linestyles = AutoProperty[tuple[Literal["solid", "dashed", "dashdot", "dotted"]]](allow_uninitialised = True)
    alpha_values = AutoProperty[tuple["float|ArrayLike"]](allow_uninitialised = True)
    colours = AutoProperty[tuple[ColorType]](allow_uninitialised = True)
    def __init__(self, **kwargs):
        super().__init__("x", "y", "z", "levels", "linewidths", "linestyles", "alpha_values", "colours", **kwargs)
    def render(self, figure: Figure, axis: Axes, *args: Any, **kwargs: Any) -> None:
        self._result = axis.contour(
            self.x,
            self.y,
            self.z,
            levels = self.levels,
            linewidths = self.linewidths,
            linestyles = self.linestyles,
            alpha = self.alpha_values,
            colors = self.colours,
            **kwargs
        )
    @staticmethod
    def from_contours(contours: QuadContourSet, x: np.ndarray[tuple[int], np.dtype[np.floating]], y: np.ndarray[tuple[int], np.dtype[np.floating]], z: np.ndarray[tuple[int], np.dtype[np.floating]], uses_single_colour_value: bool = False) -> "CachedPlotContour":
        return CachedPlotContour(
            x = x,
            y = y,
            z = z,
            levels = tuple(contours.levels),
            linewidths = contours.get_linewidths(),
            linestyles = tuple(contours.linestyles),
            alpha_values = contours.alpha,
            colours = contours.colors if not uses_single_colour_value else tuple([contours.colors] * len(contours.levels)),
            _result = contours
        )



from abc import ABC, abstractmethod
import copy
from typing import List, Tuple, Dict, Union

class _Mapping(object):
    def __init__(self):
        self.__writable = True
        self.__data = {}
        self.__sub_mapping_keys = []

    def __ensure_writable(func):
        def wrapper(self, *args, **kwargs):
            if self.__writable:
                return func(self, *args, **kwargs)
            else:
                raise RuntimeError("Unable to write to a non-writable configuration map.")
        return wrapper

    def __internal_get_item(self, *keys: List[str]):
        if keys[0] in self.__data:
            value = self.__data[keys[0]]
            if len(keys) > 1:
                if isinstance(value, _Mapping):
                    return value[keys[1:]]
                else:
                    raise TypeError(f"Value of {keys[0]} is a root node and cannot be automatically indexed further.")
            else:
                return value
        else:
            raise AttributeError(f"{keys[0]} is not a valid attribute or config name.")

    def __getitem__(self, *keys: List[str]):
        #return copy.copy(self.__internal_get_item(*keys))
        return self.__internal_get_item(*keys)

    #@__ensure_writable
    #def __setitem__(self, key: str, value):
    #    print("HERE")

    def __getattr__(self, key: str):
        #return copy.copy(self.__internal_get_item(key))
        return self.__internal_get_item(key)

    #def __setattr__(self, key: str, value):
    #    super().__setattr__(key, value)
    #    if key != "_Mapping__data" and key in self.__data:
    #        self[key] = value

    @__ensure_writable
    def add_item(self, name: str, value):
        self.__data[name] = value

    @__ensure_writable
    def add_group(self, name: str):
        self.__data[name] = _Mapping()
        self.__sub_mapping_keys.append(name)

    def lock(self):
        self.__writable = False
        for key in self.__sub_mapping_keys:
            self.__data[key].lock()

    @property
    def writable(self):
        return self.__writable

    def _get_keys(self) -> Tuple[Union[str, Dict[str, Tuple]]]:
        print("WARNING: this method is depreciated and will be removed in a future update! Use object.keys instead.")
        #return ((self.__data[key]._get_keys() if key in self.__sub_mapping_keys else key) for key in self.__data)
        return list(self.__data.keys())

    @property
    def keys(self):
        return list(self.__data.keys())

class ConfigsBase(_Mapping, ABC):
    def __init__(self, writable: bool = True, filepath: str = None):
        self.__stringify_storage = None
        self.__stringify_indent = None

        super().__init__()

        if filepath is not None:
            self.read(filepath)

        if not writable:
            self.lock()

    @abstractmethod
    def read(self, filepath: str):
        raise NotImplementedError()

    @staticmethod
    @abstractmethod
    def from_file(filepath: str, writable: bool = False, *args, **kwargs):
        raise NotImplementedError()

    def __str__(self):
        highrarchy = self

        self.__stringify_storage = ""
        self.__stringify_indent = 0
        def recursive_writer(sub_highrarchy):
            all_keys = sub_highrarchy.keys

            if len(all_keys) == 0:
                return# There is nothing to display here

            non_mapping_key_indexes = [i for i, key in enumerate(all_keys) if not isinstance(sub_highrarchy[key], _Mapping)]
            if len(non_mapping_key_indexes) > 0:
                max_key_length = max([len(all_keys[i]) for i in non_mapping_key_indexes])

                # Display non-mapables
                for key_index in non_mapping_key_indexes:
                    key = all_keys[key_index]
                    value = sub_highrarchy[key]

                    insert_value = ('"'+value+'"') if isinstance(value, str) else value
                    self.__stringify_storage += (" " * self.__stringify_indent) + f"{key}{' ' * (max_key_length - len(key))} = {insert_value}\n"

            # Recursivley display mappables
            for key in (all_keys if len(non_mapping_key_indexes) == 0 else [key for i, key in enumerate(all_keys) if i not in non_mapping_key_indexes]):
                value = sub_highrarchy[key]

                self.__stringify_storage += (" " * self.__stringify_indent) + "\n"
                self.__stringify_storage += (" " * self.__stringify_indent) + f"{key}:\n"
                self.__stringify_indent += 4
                recursive_writer(value)
                self.__stringify_indent -= 4

        recursive_writer(highrarchy)
        return self.__stringify_storage.lstrip("\n")

    def __repr__(self):
        return "Configuration Object. Items as follows:\n\n" + self.__str__()



from typing import Generic, TypeVar, Union, Iterable, Callable, Any, cast

from ._cast import Cast
from ._typeshield import TypeShield_Base

T = TypeVar("T")

class AutoProperty(property, Generic[T]):
    """
    Automation of the property type that automatically handles the creation of a hidden attribute.

    Uninitialised properties are assigned None and are considered uninitialised (i.e. like NULLPTR).

    Calling the deleter will uninitialise the value back to None.

    Constructor:
        str | None doc                 -> Optional documentation string for the property. Defaults to None.
              bool allow_uninitialised -> If the property is unititialised, should None be returned upon attempting to access the value? Default is False.
    """
    
    def __init__(self, default_value: Union[T, None] = None, doc: Union[str, None] = None, allow_uninitialised: bool = False) -> None:
        self.__name: Union[str, None] = None
        self.__declaring_type_name: Union[str, None] = None
        self.__check_uninitialised: bool = not (allow_uninitialised or default_value is not None)
        self.__default_value = default_value
        super().__init__(doc = doc)

    def __set_name__(self, owner, name):
        self.__declaring_type_name = owner.__qualname__
        self.__name = name

    @property
    def _name(self) -> Union[str, None]:
        return self.__name
    
    def is_initialised(self, instance: Any) -> Union[bool, None]:
        if self.__check_uninitialised:
            initilised: Union[bool, None] = None
            try:
                initilised = cast(bool, getattr(instance, self.__get_storage_attribute_name() + "__isinit"))
            except AttributeError:
                initilised = False
                setattr(instance, self.__get_storage_attribute_name() + "__isinit", False)
            return initilised
        else:
            return None

    def __get_storage_attribute_name(self):
        return f"_{self.__declaring_type_name}__autoprop_{self.__name}"

    def __get__(self, instance: Any, owner: Union[type, None] = None, /) -> Union[T, None]:
        if self.__check_uninitialised:
            if not self.is_initialised(instance):
                raise ValueError("Attempted to access value of uninitialised AutoProperty with initialisation enforcement enabled.")

        try:
            return getattr(instance, self.__get_storage_attribute_name())
        except AttributeError:
            setattr(instance, self.__get_storage_attribute_name(), self.__default_value)
            return getattr(instance, self.__get_storage_attribute_name())

    def __set__(self, instance: Any, value: Any, /) -> None:
        setattr(instance, self.__get_storage_attribute_name(), value)
        if self.__check_uninitialised:
            setattr(instance, self.__get_storage_attribute_name() + "__isinit", True)

    def __delete__(self, instance: Any, /) -> None:
        setattr(instance, self.__get_storage_attribute_name(), None)
        if self.__check_uninitialised:
            setattr(instance, self.__get_storage_attribute_name() + "__isinit", False)

class AutoProperty_NonNullable(AutoProperty[T]):
    """
    Automation of the property type that automatically handles the creation of a hidden attribute.
    A value must be assigned before attempting to read from a property of this type.
    Attempting to read from an uninitialised property will result in a ValueError.

    Syntactic sugar for AutoProperty(..., allow_uninitialised = False)
    Also adds a more appropriate type hint for the underlying property __get__ method.

    Calling the deleter will uninitialise the value making the property unreadable.

    Constructor:
        str | None doc -> Optional documentation string for the property. Defaults to None.
    """

    def __init__(self, default_value: Union[T, None] = None, doc: Union[str, None] = None) -> None:
        super().__init__(default_value = default_value, doc = doc, allow_uninitialised = False)

    def __get__(self, instance: Any, owner: Union[type, None] = None, /) -> T:
        return cast(T, super().__get__(instance, owner))

class NullableTypedAutoProperty(AutoProperty[T]):
    """
    Automation of the property type that automatically handles the creation of a hidden attribute.

    Uninitialised properties are assigned None by default.

    Calling the deleter will uninitialise the value back to None.

    TypeParam:
        T -> The type of the property's value (when initialised).

    Constructor:
        TypeShield_Base type_shield -> Type objects (and None) that are considered valid types for the property's value.
             str | None doc         -> Optional documentation string for the property. Defaults to None.
    """

    def __init__(self, type_shield: TypeShield_Base, default_value: Union[T, None] = None, doc: Union[str, None] = None) -> None:
        self.__valid_type_definition_shield = type_shield
        super().__init__(default_value = default_value, doc = doc, allow_uninitialised = True)

    def __set__(self, instance: Any, value: Any, /) -> None:
        super().__set__(instance, self.__valid_type_definition_shield.shield_check(value) if value is not None else None)

class TypedAutoProperty(AutoProperty_NonNullable[T]):
    """
    Automation of the property type that automatically handles the creation of a hidden attribute.

    Uninitialised properties are assigned None and are considered uninitialised (i.e. like NULLPTR).

    Calling the deleter will uninitialise the value back to None.

    TypeParam:
        T -> The type of the property's value (when initialised).

    Constructor:
        TypeShield_Base type_shield -> Type objects (and None) that are considered valid types for the property's value.
             str | None doc         -> Optional documentation string for the property. Defaults to None.
    """

    def __init__(self, type_shield: TypeShield_Base, default_value: Union[T, None] = None, doc: Union[str, None] = None) -> None:
        self.__valid_type_definition_shield = type_shield
        super().__init__(default_value = default_value, doc = doc)

    def __set__(self, instance: Any, value: Any, /) -> None:
        super().__set__(instance, self.__valid_type_definition_shield.shield_check(value))

class TypeCastAutoProperty(AutoProperty[T]):
    """
    Automation of the property type that automatically handles the creation of a hidden attribute.

    Uninitialised properties are assigned None and are considered uninitialised (i.e. like NULLPTR).

    Calling the deleter will uninitialise the value back to None.

    TypeParam:
        T -> The return type of the cast operation. Also the type of the property's value (when initialised).

    Constructor:
        str | None doc                 -> Optional documentation string for the property. Defaults to None.
              bool allow_uninitialised -> If the property is unititialised, should None be returned upon attempting to access the value? Default is False.
    """

    def __init__(self, cast: Union[Cast[T], Callable[[Any], T]], default_value: Union[T, None] = None, doc: Union[str, None] = None) -> None:
        self.__cast: Union[Cast[T], Callable[[Any], T]] = cast
        super().__init__(default_value = default_value, doc = doc, allow_uninitialised = cast.nullable if isinstance(cast, Cast) else False)

    def __set__(self, instance: Any, value: Any, /) -> None:
        super().__set__(instance, self.__cast(value))



import pytest

from QuasarCode.Games.Spinners import Spinner, ISpinner
from QuasarCode.Games.Dice import NDice, IDice, Dice6, Dice8, Dice12

class TestClass_Games(object):
    def test_Spinners(self):
        spinner = Spinner(["Red", "Yellow", "Blue", "Green"])

        for i in range(100):
            print(spinner.Spin())

    def test_Dice(self):
        dice = NDice(6)

        for i in range(100):
            print(dice.Roll())
        print()
        
        assert len(NDice.MultipleDice([6, 6])) == 2, "Wrong number of dice produced by MultipleDice."
        assert len(NDice.MultipleDice([6], noOfDice = 2)) == 2, "Wrong number of dice produced by MultipleDice."

        dice6 = Dice6()
        dice8 = Dice8()
        dice12 = Dice12()

        print(dice6.Roll())

        print(dice8.Roll())

        print(dice12.Roll())

    def test_Cards(self):
        pass#TODO:



#from ._convinience_methods import synchronyse, mpi_check_equal, if_mpi_root, mpi_barrier, mpi_get_slice, mpi_slice, mpi_gather_array, mpi_scatter_array, mpi_redistribute_array_evenly
from ._convinience_methods import synchronyse, mpi_check_equal, if_mpi_root, mpi_barrier, mpi_sum, \
    mpi_mean, mpi_get_slice, mpi_slice, mpi_gather_array, mpi_scatter_array, \
    mpi_redistribute_array_evenly, mpi_argsort, mpi_sort, mpi_calculate_reorder, mpi_reorder, \
    mpi_apply_reorder



from decimal import Decimal
from typing import Iterable

import numpy as np


class Rect:
    """
    A 2D rectangle.

    Parameters:
        float x_min:
            X-coordinate of the bottom-left corner.
        float x_max:
            X-coordinate of the top-right corner.
        float y_min:
            Y-coordinate of the bottom-left corner.
        float y_max:
            Y-coordinate of the top-right corner.

    Use the following static methods to create an instance:
        - create_from_size(x, y, width, height)
        - create_from_center(x, y, width, height)
        - create_from_limits(x_min, x_max, y_min, y_max)
    """
    def __init__(self, x_min: float, x_max: float, y_min: float, y_max: float):
        self.__x_min = x_min
        self.__x_max = x_max
        self.__y_min = y_min
        self.__y_max = y_max

    @staticmethod
    def create_from_size(x: float, y: float, width: float, height: float) -> "Rect":
        """
        Creates a Rect instance using the specified position of the bottom-left corner and size.

        Parameters:
            float x:
                The x-coordinate of the bottom-left corner of the rectangle.
            float y:
                The y-coordinate of the bottom-left corner of the rectangle.
            float width:
                The width of the rectangle.
            float height:
                The height of the rectangle.

        Returns:
            Rect -> A new Rect instance defined by the given position and size.
        """
        return Rect(x, x + width, y, y + height)

    @staticmethod
    def create_from_center(x: float, y: float, width: float, height: float) -> "Rect":
        """
        Create a Rect instance using the center coordinates and dimensions.

        Parameters:
            float x:
                The x-coordinate of the center of the rectangle.
            float y:
                The y-coordinate of the center of the rectangle.
            float width:
                The width of the rectangle.
            float height:
                The height of the rectangle.

        Returns:
            Rect -> Instance centred on (x, y) with the specified width and height.
        """
        return Rect(x - width / 2, x + width / 2, y - height / 2, y + height / 2)

    @staticmethod
    def create_from_limits(x_min: float, x_max: float, y_min: float, y_max: float) -> "Rect":
        """
        Create a Rect instance from the specified boundary limits.
        This is the same as the constructor parameters and is implemented for consistency.

        Parameters:
            float x_min:
                The minimum x-coordinate of the rectangle.
            float x_max:
                The maximum x-coordinate of the rectangle.
            float y_min:
                The minimum y-coordinate of the rectangle.
            float y_max:
                The maximum y-coordinate of the rectangle.

        Returns:
            Rect -> A new instance of the Rect class defined by the given limits.
        """
        return Rect(x_min, x_max, y_min, y_max)

    @staticmethod
    def create_from_data(x_data: Iterable[float], y_data: Iterable[float]) -> "Rect":
        """
        Create a Rect instance that encompasses the provided data.

        Parameters:
            Iterable[float] x_data:
                X-coordinates.
            Iterable[float] y_data:
                Y-coordinates.

        Returns:
            Rect -> A new instance of the Rect class that encompasses the data.
        """
        return Rect(min(x_data), max(x_data), min(y_data), max(y_data))

    @property
    def extent(self) -> tuple[float, float, float, float]:
        """
        The limits of the rectangle in the form (x_min, x_max, y_min, y_max).
        This corresponds to the format of a matplotlib 'extent' parameter.
        """
        return (self.__x_min, self.__x_max, self.__y_min, self.__y_max)

    @property
    def range(self) -> tuple[tuple[float, float], tuple[float, float]]:
        """
        The limits of the rectangle in the form ((x_min, x_max), (y_min, y_max)).
        This corresponds to the format of a numpy histogram2d 'range' parameter.
        """
        return ((self.__x_min, self.__x_max), (self.__y_min, self.__y_max))

    @property
    def width(self) -> float:
        """
        The width of the rectangle.

        Returns:
            float -> The width of the rectangle.
        """
        return self.__x_max - self.__x_min

    @property
    def height(self) -> float:
        """
        The height of the rectangle.
        
        Returns:
            float -> The height of the rectangle.
        """
        return self.__y_max - self.__y_min

    @property
    def centre(self) -> tuple[float, float]:
        """
        The coordinates of the center of the rectangle.

        Returns:
            tuple[float, float] -> (x, y)
        """
        return ((self.__x_max + self.__x_min) / 2, (self.__y_max + self.__y_min) / 2)

    @property
    def area(self) -> float:
        """
        The area of the rectangle.

        Returns:
            float -> The area of the rectangle.
        """
        return self.width * self.height

    def __copy__(self) -> "Rect":
        """
        Create a shallow copy of the rectangle.

        Returns:
            Rect -> A new instance of the Rect class with the same properties.
        """
        return Rect(self.__x_min, self.__x_max, self.__y_min, self.__y_max)
    
    def __str__(self) -> str:
        return f"Rect(x: {self.__x_min} -> {self.__x_max}, y: {self.__y_min} -> {self.__y_max})"
    
    def __repr__(self) -> str:
        return self.__str__()
    
    def __eq__(self, value):
        if isinstance(value, Rect):
            return self.__x_min == value.__x_min and self.__x_max == value.__x_max and self.__y_min == value.__y_min and self.__y_max == value.__y_max
        return False
    
    def __ne__(self, value):
        return not self.__eq__(value)
    
    def __gt__(self, value):
        if isinstance(value, Rect):
            return self.area > value.area
        elif isinstance(value, (int, float, Decimal, np.integer, np.floating)):
            return self.area > value
        return False
    
    def __ge__(self, value):
        if isinstance(value, Rect):
            return self.area >= value.area
        elif isinstance(value, (int, float, Decimal, np.integer, np.floating)):
            return self.area >= value
        return False
    
    def __lt__(self, value):
        if isinstance(value, Rect):
            return self.area < value.area
        elif isinstance(value, (int, float, Decimal, np.integer, np.floating)):
            return self.area < value
        return False
    
    def __le__(self, value):
        if isinstance(value, Rect):
            return self.area <= value.area
        elif isinstance(value, (int, float, Decimal, np.integer, np.floating)):
            return self.area <= value
        return False



from matplotlib.figure import Figure
from matplotlib.axes import Axes

from ..Data._Rect import Rect
from ..Tools._Struct import CacheableStruct
from ..Tools._autoproperty import AutoProperty, AutoProperty_NonNullable
from ._CachedPlotElements import CachedPlotElement, CachedPlotColourbar

class CachedPlot(CacheableStruct):
    plot_elements = AutoProperty_NonNullable[dict[str, CachedPlotElement]]()
    colourbars = AutoProperty_NonNullable[dict[str, CachedPlotColourbar]]()
    extent = AutoProperty_NonNullable[Rect]()
    x_axis_label = AutoProperty[str](allow_uninitialised = True)
    y_axis_label = AutoProperty[str](allow_uninitialised = True)
    flip_x = AutoProperty_NonNullable[bool](default_value = False)
    flip_y = AutoProperty_NonNullable[bool](default_value = False)
    show_legend = AutoProperty_NonNullable[bool](default_value = False)
    def __init__(self, **kwargs):
        super().__init__(
            cacheable_attributes = ("plot_elements", "colourbars", "extent", "x_axis_label", "y_axis_label", "flip_x", "flip_y", "show_legend"),
            **kwargs
        )
        self.plot_elements = {}
        self.colourbars = {}
    
    def __check_element_name_is_new(self, name: str) -> bool:
        return name not in self.plot_elements

    def add_element(self, name: str, element: CachedPlotElement) -> None:
        if not self.__check_element_name_is_new(name):
            raise KeyError(f"An element already exists with the name \"{name}\".")
        self.plot_elements[name] = element

    def add_colourbar(self, name: str, colourbar: CachedPlotColourbar) -> None:
        if name in self.colourbars:
            raise KeyError(f"A colourbar already exists with the name \"{name}\".")
        if not self.__check_element_name_is_new(colourbar.target_element):
            raise KeyError(f"No element exists with the target name \"{colourbar.target_element}\".")
        self.colourbars[name] = colourbar

    def render(self, figure: Figure, axis: Axes):
        for element in self.plot_elements:#TODO: why can this be __has_cacheables after loading from file!!!!!
            self.plot_elements[element].render(figure, axis)

        for name, colourbar in self.colourbars.items():
            target: object
            if colourbar.target_element in self.plot_elements:
                target = self.plot_elements[colourbar.target_element]._result
            else:
                raise RuntimeError(f"Unable to locate target for colourbar \"{name}\".")
            colourbar.render(figure, axis, target)

        if self.x_axis_label is not None:
            axis.set_xlabel(self.x_axis_label)
        if self.y_axis_label is not None:
            axis.set_ylabel(self.y_axis_label)
        
        axis.set_xlim(self.extent.extent[0:2] if not self.flip_x else self.extent.extent[1:-1:-1])
        axis.set_ylim(self.extent.extent[2:4] if not self.flip_y else self.extent.extent[3:1:-1])

        if self.show_legend:
            axis.legend()



from ._Bins import Bins
from ._CachedPlotElements import \
    CachedPlotElement,  \
    CachedPlotLine,     \
    CachedPlotScatter,  \
    CachedPlotErrorbar, \
    CachedPlotHexbin,   \
    CachedPlotContour,  \
    CachedPlotColourbar
from ._CachedPlot import CachedPlot
from ._CachedPlotFactory import CachedPlotFactory
from ._Hexbin import Hexbin
from ._Contour import Contour



from abc import abstractmethod
from typing import Union, Collection, TypeVar, Type, Any

from ..IO.Caching._Cacheable import Cacheable

class Struct(object):
    """
    Structure type definition.
    Allows a number of public instance attributes to be declared and populated from the constructor.

    Constructor Design Order:
        1.) Declare all public attributes first.
        2.) super().__init__(**kwargs)

    To specify a human readable type name for errors, overload: _get_typename(self) -> Union[str, None]
    To specify type casts for attributes, set fields to an instance of TypeCastAutoProperty
    """
    
    def __init__(self, **kwargs):
        for key, value in kwargs.items():
            try:
                setattr(self, key, value)
            except AttributeError as e:
                raise AttributeError(f"{self._get_typename if self._get_typename is not None else type(self).__qualname__} struct defines no public field named \"{key}\".") from e

    @staticmethod
    def _get_typename() -> Union[str, None]:
        """
        Overload this to specify a string to appear in the error message provided when attempting to initalise a nonexistant attribute.
        """
        return None



T = TypeVar("T", bound = "CacheableStruct")

class CacheableStruct(Struct, Cacheable):
    """
    Structure type definition.
    Allows a number of public instance attributes to be declared and populated from the constructor.

    Constructor Design Order:
        1.) Declare all public attributes first.
        2.) super().__init__(cacheable_attributes = [<attribute-names>], **kwargs)
            where <attribute-names> is a list of the attributes that should be cached.

    To specify a human readable type name for errors, overload: _get_typename(self) -> Union[str, None]
    To specify type casts for attributes, set fields to an instance of TypeCastAutoProperty
    """
    
    def __init__(self, cacheable_attributes: Collection[str], **kwargs):
        self.__cacheable_attributes: tuple[str, ...] = tuple(cacheable_attributes)
        super().__init__(**kwargs)

    @classmethod
    def __from_cache_data__(cls: Type[T], data: dict[str, Any]) -> T:
        cacheable_attributes = data.pop("cacheable_attributes", None)
        instance = cls()
        for attribute_name in data:
            if attribute_name == "__has_cacheables":
                continue
            setattr(instance, attribute_name, data[attribute_name] if not isinstance(data[attribute_name], dict) or "datatype" not in data[attribute_name] else data[attribute_name]["datatype"].__from_cache_data__(data[attribute_name]["data"]))
            if isinstance(getattr(instance, attribute_name), dict) and "__has_cacheables" in getattr(instance, attribute_name):
                for sub_key in getattr(instance, attribute_name):
                    if sub_key == "__has_cacheables":
                        continue
                    elif isinstance(getattr(instance, attribute_name)[sub_key], dict) and "datatype" in getattr(instance, attribute_name)[sub_key]:
                        getattr(instance, attribute_name)[sub_key] = getattr(instance, attribute_name)[sub_key]["datatype"].__from_cache_data__(getattr(instance, attribute_name)[sub_key]["data"])
                getattr(instance, attribute_name).pop("__has_cacheables")
        return instance

    def __get_cache_data__(self) -> dict[str, Any]:
        data = {
            attribute_name : getattr(self, attribute_name)
            for attribute_name
            in self.__cacheable_attributes
        }
        for key in data:
            if isinstance(data[key], Cacheable):
                data[key] = { "datatype" : type(data[key]), "data" : data[key].__get_cache_data__() }
            elif isinstance(data[key], dict):
                contains_cacheable = False
                for sub_key in data[key]:
                    if isinstance(data[key][sub_key], Cacheable):
                        contains_cacheable = True
                        data[key][sub_key] = { "datatype" : type(data[key][sub_key]), "data" : data[key][sub_key].__get_cache_data__() }
                if contains_cacheable:
                    data[key]["__has_cacheables"] = True
        return data



from collections.abc import Callable, Iterable, Sequence
import ctypes
from typing import TypeVar, ParamSpec
from functools import wraps
import inspect
import sys



P = ParamSpec("P")
T = TypeVar("T")



def update_locals(*local_variable_parameter_names_or_indexes: str|int, remove_target_arguments: bool = True):
    if len(local_variable_parameter_names_or_indexes) == 0:
        raise IndexError("No local variable parameter information provided.")

    def inner(func: Callable[P, T]) -> Callable[P, None]:
        @wraps(func)
        def wrapper(*args, **kwargs):
            args = list(args) # Make args mutable

            SKIP_THIS_FRAME = True # Adds a local variable that can be checked for in case of stacked calls
            frame = inspect.currentframe()
            while True:
                frame = frame.f_back
                if frame is None:
                    raise RuntimeError("Unable to find root calling frame. A valid calling frame must not have a local named \"SKIP_THIS_FRAME\".")
                calling_scope_locals = frame.f_locals
                if "SKIP_THIS_FRAME" not in calling_scope_locals:
                    break

            target_variable_names = []
            removable_args_indexes = []
            removable_kwargs_keys = []
            for name_param_location in local_variable_parameter_names_or_indexes:
                if isinstance(name_param_location, int):
                    removable_args_indexes.append(name_param_location)
                    target_variable_names.append(args[name_param_location])
                else:
                    removable_kwargs_keys.append(name_param_location)
                    target_variable_names.append(kwargs[name_param_location])
            if remove_target_arguments:
                removable_args_indexes.sort(reverse = True)
                for i in removable_args_indexes:
                    args.pop(i)
                for key in removable_kwargs_keys:
                    kwargs.pop(key)

            result = func(*args, **kwargs)

            if len(target_variable_names) == 1:
                calling_scope_locals[target_variable_names[0]] = result

            else:
                if not isinstance(result, Iterable):
                    raise TypeError(f"Decorated method return type of {type(result)} is not Iterable but multiple target varaibles were specified. Unable to unpack non-iterable output.")
                result = tuple(result)
                if len(result) != len(target_variable_names):
                    raise IndexError(f"Unable to unpack {len(result)} values into {len(target_variable_names)} variables.")
                for i in range(len(target_variable_names)):
                    calling_scope_locals[target_variable_names[i]] = result[i]

            if sys.version_info < (3, 13):
                ctypes.pythonapi.PyFrame_LocalsToFast(ctypes.py_object(frame), ctypes.c_int(0))

        return wrapper
    return inner



def read_locals(*local_variable_parameter_names_or_indexes: str|int, treat_unititilised_as_none: bool = True):

    if len(local_variable_parameter_names_or_indexes) == 0:
        def inner(func: Callable[P, T]) -> Callable[P, T]:
            return func
        return inner

    def inner(func: Callable[P, T]) -> Callable[..., T]:
        @wraps(func)
        def wrapper(*args, **kwargs):
            args = list(args) # Make args mutable

            SKIP_THIS_FRAME = True # Adds a local variable that can be checked for in case of stacked calls
            frame = inspect.currentframe()
            while True:
                frame = frame.f_back
                if frame is None:
                    raise RuntimeError("Unable to find root calling frame. A valid calling frame must not have a local named \"SKIP_THIS_FRAME\".")
                calling_scope_locals = frame.f_locals
                if "SKIP_THIS_FRAME" not in calling_scope_locals:
                    break

            for name_param_location in local_variable_parameter_names_or_indexes:
                var_name: str
                if isinstance(name_param_location, int):
                    var_name = args[name_param_location]
                else:
                    var_name = kwargs[name_param_location]
                if var_name not in calling_scope_locals:
                    if treat_unititilised_as_none:
                        args[name_param_location] = None
                    else:
                        raise NameError(f"No initialised variable with the name \"{var_name}\".")
                else:
                    args[name_param_location] = calling_scope_locals[var_name]

            return func(*args, **kwargs)

        return wrapper
    return inner



def use_locals(readable_parameter_names_or_indexes: str|int|Sequence[str|int], updateable_parameter_names_or_indexes: str|int|Sequence[str|int], remove_extra_updateable_target_arguments: bool = True, treat_unititilised_as_none: bool = True):
    if not isinstance(readable_parameter_names_or_indexes, Sequence):
        readable_parameter_names_or_indexes = (readable_parameter_names_or_indexes, )
    if not isinstance(updateable_parameter_names_or_indexes, Sequence):
        updateable_parameter_names_or_indexes = (updateable_parameter_names_or_indexes, )
    
    if len(updateable_parameter_names_or_indexes) == 0:
        raise IndexError("No updateable local variable parameter information provided.")

    def inner(func: Callable[P, T]) -> Callable[P, None]:
        @wraps(func)
        def wrapper(*args, **kwargs):
            args = list(args) # Make args mutable

            SKIP_THIS_FRAME = True # Adds a local variable that can be checked for in case of stacked calls
            frame = inspect.currentframe()
            while True:
                frame = frame.f_back
                if frame is None:
                    raise RuntimeError("Unable to find root calling frame. A valid calling frame must not have a local named \"SKIP_THIS_FRAME\".")
                calling_scope_locals = frame.f_locals
                if "SKIP_THIS_FRAME" not in calling_scope_locals:
                    break

            target_variable_names = []
            removable_args_indexes = []
            removable_kwargs_keys = []
            for name_param_location in updateable_parameter_names_or_indexes:
                if isinstance(name_param_location, int):
                    if name_param_location not in readable_parameter_names_or_indexes:
                        removable_args_indexes.append(name_param_location)
                    target_variable_names.append(args[name_param_location])
                else:
                    if name_param_location not in readable_parameter_names_or_indexes:
                        removable_kwargs_keys.append(name_param_location)
                    target_variable_names.append(kwargs[name_param_location])

            for name_param_location in readable_parameter_names_or_indexes:
                if isinstance(name_param_location, int):
                    var_name = args[name_param_location]
                else:
                    var_name = kwargs[name_param_location]
                if var_name not in calling_scope_locals:
                    if treat_unititilised_as_none:
                        args[name_param_location] = None
                    else:
                        raise NameError(f"No initialised variable with the name \"{var_name}\".")
                else:
                    args[name_param_location] = calling_scope_locals[var_name]

            if remove_extra_updateable_target_arguments:
                removable_args_indexes.sort(reverse = True)
                for i in removable_args_indexes:
                    args.pop(i)
                for key in removable_kwargs_keys:
                    kwargs.pop(key)

            result = func(*args, **kwargs)

            if len(target_variable_names) == 1:
                calling_scope_locals[target_variable_names[0]] = result

            else:
                if not isinstance(result, Iterable):
                    raise TypeError(f"Decorated method return type of {type(result)} is not Iterable but multiple target varaibles were specified. Unable to unpack non-iterable output.")
                result = tuple(result)
                if len(result) != len(target_variable_names):
                    raise IndexError(f"Unable to unpack {len(result)} values into {len(target_variable_names)} variables.")
                for i in range(len(target_variable_names)):
                    calling_scope_locals[target_variable_names[i]] = result[i]

            if sys.version_info < (3, 13):
                ctypes.pythonapi.PyFrame_LocalsToFast(ctypes.py_object(frame), ctypes.c_int(0))

        return wrapper
    return inner



import numpy as np
from matplotlib import pyplot as plt

from .figure import Figure

class Graph(object):
    """
    Displays multiple figures as a single figure and allows them to be saved as a single figure.

    Arguments:
        list[Figure] figures -> A list of figures to display in order. Must have at least one item if the "data" paramiter is blank. (Default is [])

        list[list] data -> A list of lists of paramiters to be passed to the Figure object constructor. Must have at least one item if the "figures" paramiter is blank. (Default is [])

        tuple shape -> The number of columns and rows to use to display the individual figures. (Default is (1, 1))

        tuple size -> A tuple with the horizontal and vertical size of each individual figure. (Defult is (6.4, 4.8))

        string title -> The title to be displayed on the graph. (Default is "")

        string saveName -> The png file name (can be precided with a file path) to save the graph under. (Default is "figure.png")
    """

    def __init__(self, figures = [], data = [], shape = (1, 1), size = None, title = "", saveName = "figure.png"):
        self.title = title
        self.saveName = saveName
        self.size = size if size is not None else [6.4, 4.8]# This is the deafult as documented by https://matplotlib.org/3.1.1/api/_as_gen/matplotlib.pyplot.figure.html

        self.__shape = shape if len(shape) > 1 else (shape[0], 1)
        self.figure, self.axes = plt.subplots(ncols = self.__shape[0], nrows = self.__shape[1])
        if not isinstance(self.axes, np.ndarray):
            self.axes = [[self.axes]]
        elif not isinstance(self.axes[0], np.ndarray):
            self.axes = [self.axes]

        self.figures = figures

        self.data = np.array(data)

        # The number of extra figures needed to hold the data provided
        requiredFigures = np.shape(self.data)[0]#TODO: make this a more robust check

        if requiredFigures + len(self.figures) > self.__shape[0] * self.__shape[1]:
            raise ValueError("The shape provided didn't allow enough space for all of the figures provided. The shape allowed for {} but {} were avalable.".format(shape[0] * (shape[1] if len(shape) > 1 else 1), requiredFigures + len(self.figures)))

        # Adds the extra figures
        for i in range(requiredFigures):
            self.figures.append(Figure(*self.data[i]))

        self.numberOfFigures = len(self.figures)

        self.plot()

    def plot(self, options = None):
        """
        Draws the figure onto an axis.

        Arguments:
            matplotlib.axes._subplots.AxesSubplot axisOveride -> an alternitive axis to draw on.

            dict options -> Named arguments to be passed to all the plot methods of the contained figure objects.
                            Argument names should be strings. (Default is None)

            OR

            list[dict] options -> A list of dicts - one for each figure for the same purpose as the above. (Default is None)
        """
        for i in range(self.numberOfFigures):
            axis = self.axes[i // self.__shape[0]][i - (i // self.__shape[1]) * self.__shape[0]]
            axis.cla()
            self.figures[i].plot(axis, **(options if isinstance(options, dict) else (options[i] if isinstance(options, list) and len(options) > i and isinstance(options[i], dict) else {}) ))

        self.figure.set_figwidth(self.size[0])
        self.figure.set_figheight(self.size[1])
            
    def save(self, nameOveride = None):
        """
        Saves the graph in its current state as a png using the current save name.

        Arguments:
            string nameOveride -> The filename or filepath to use instead of the one stored by the object.
        """
        if nameOveride is not None and nameOveride != "":
            self.figure.savefig(nameOveride)
        elif self.saveName is not None and self.saveName != "":
            self.figure.savefig(self.saveName)

    def show(self):
        """
        Show the figure(s) in a window.
        """
        self.figure.show()


import os

class _Settings(object):
    __singleton = None

    @staticmethod
    def _is_singleton_avalible():
        return _Settings.__singleton is not None

    @staticmethod
    def _get_singleton():
        return _Settings.__singleton

    __attribute_initial_values = {
                       "verbose" : False,
               "verbosity_level" : -1,
                         "debug" : False,
                  "mpi_avalible" : False,
                         "slurm" : "SLURM_JOB_ID" in os.environ,
               "datetime_format" : r"%d/%m/%Y, %H:%M:%S",
                   "date_format" : r"%d/%m/%Y",
                   "time_format" : r"%H:%M:%S",
           "time_format_precise" : "%H:%M:%S[%f\u03BCs]",
        "cuda_threads_per_block" : 512,
    }

    def __init__(self):
        if not _Settings._is_singleton_avalible():
            _Settings.__singleton = self
            self.__setting_values = _Settings.__attribute_initial_values.copy()
            self.__verbosity_level_cache: int = self.__setting_values["verbosity_level"]

        else:
            raise RuntimeError("Only one instance of the __Settings object may exist.")

    def __getattr__(self, name: str):
        if name in self.__setting_values:
            return self.__setting_values[name]
        else:
            raise AttributeError(f"{name} is not a valid attribute or setting name.")

    @property
    def setting_names(self):
        return list(self.__setting_values.keys())

    def _set_verbose(self, state: bool, level: int|None):
        self.__setting_values["verbose"] = state
        if state:
            if level is None:
                self.__setting_values["verbosity_level"] = 0
            else:
                self.__setting_values["verbosity_level"] = level
        else:
            self.__setting_values["verbosity_level"] = -1
    def enable_verbose(self, level: int = 0):
        self._set_verbose(True, level)
    def disable_verbose(self):
        if self.verbose:
            self.__verbosity_level_cache = self.verbosity_level
        self._set_verbose(False)
    def toggle_verbose(self):
        if self.verbose:
            self.__verbosity_level_cache = self.verbosity_level
        self._set_verbose(not self.verbose, level = self.__verbosity_level_cache)

    def _set_debug(self, state: bool): self.__setting_values["debug"] = state
    def enable_debug(self): self._set_debug(True)
    def disable_debug(self): self._set_debug(False)
    def toggle_debug(self): self._set_debug(not self.debug)

    def _set_mpi_avalible(self):
        self.__setting_values["mpi_avalible"] = True

    def set_cuda_threads_per_block(self, value: int):
        self.__setting_values["cuda_threads_per_block"] = value

if not _Settings._is_singleton_avalible():
    _Settings()

settings_object = _Settings._get_singleton()



from ._IDice import IDice
import random

class NDice(IDice):
    """
    A dice with a specified number of sides

    __init__():
        int sides --> number of sides on the dice
        int seed --> seed for the random generator (deafult is None)
    """
    def __init__(self, sides: int, seed: int = None):
        self.__Generator: random.Random = random.Random() if seed == None else random.Random(seed)

        # Number of sides on the dice
        self.Sides: int = sides

    def Roll(self):
        """
        Rolls the dice
        """
        return self.__Generator.randint(1, self.Sides + 1)

    def MultipleDice(sides: list, noOfDice: int = None):
        """
        Convenience method for creating multiple dice

        Paramiters:
            list sides --> list of side numbers representing unique dice. For multiple of the same number, use a single element and alter noOfDice
            int noOfDice --> number of dice to create (deafult is length of sides list)
        """
        if noOfDice == None:
            noOfDice = len(sides)

        if len(sides) > 1:
            randomiser = random.Random()
            return [NDice(size, randomiser.randint(100000, 999999)) for size in sides]

        else:
            randomiser = random.Random()
            return [NDice(sides[0], randomiser.randint(100000, 999999)) for i in range(noOfDice)]


from ..._global_settings import settings_object as _settings_object
from ...MPI import get_mpi_rank
from ...Tools._stopwatch import IStopwatch, SimpleStopwatch



class Console(object):
    """
    Rich test IO interface for the console.

    When using MPI, process information uses the following layout order: (MPI rank, subprocess)
    """

    __stopwatch: IStopwatch = SimpleStopwatch.start_new()
    @staticmethod
    def reset_stopwatch(synchronise: IStopwatch|None = None) -> None:
        Console.__stopwatch.stop()
        Console.__stopwatch.reset()
        if synchronise is not None:
            Console.__stopwatch.synchronise_start(synchronise)
        else:
            Console.__stopwatch.start()

    __print_newline_indentation: str = " " * 10
    @staticmethod
    def __custom_newline_format(info_insert_length, s) -> str:
        return str(s).replace("\n", f"\n{Console.__print_newline_indentation}" + (" " * info_insert_length))

    @staticmethod
    def __print(prefix: str, raw_text_input: list[str], /, show_insert: bool, **kwargs) -> None:

        # Don't print if on a rank that is blocked from printing
        if not (Console.__limit_mpi_output_ranks is None or (Console.__debug_bypasses_mpi_output_restriction and _settings_object.debug) or get_mpi_rank() in Console.__limit_mpi_output_ranks):
            return

        if prefix != "":
            prefix += " "

        insert_text: str = ""
        if show_insert:
            if _settings_object.mpi_avalible or Console.__process_id is not None:
                process_list = []
                if _settings_object.mpi_avalible:
                    process_list.append(str(get_mpi_rank()))
                if Console.__process_id is not None:
                    process_list.append(str(Console.__process_id))
                insert_text += "(" + ",".join(process_list) + ") "#TODO: add padding for numbers with fewer digits than the largest number in each category
            if Console.__show_time:
                insert_text += f"[{Console.__stopwatch.get_elapsed_time():10.3f}] "
        insert_test_length = len(insert_text)

        # Indent newlines
        text: str = " ".join([Console.__custom_newline_format(insert_test_length, str(s)) for s in raw_text_input])

        if "flush" not in kwargs and (_settings_object.debug or _settings_object.slurm or _settings_object.mpi_avalible):
            kwargs["flush"] = True

        print(f"{prefix}{insert_text}{text}", **kwargs)

    __show_time: bool = False
    @staticmethod
    def show_times():
        Console.__show_time = True
    @staticmethod
    def hide_times():
        Console.__show_time = False

    __limit_mpi_output_ranks: tuple[int]|None = None
    __debug_bypasses_mpi_output_restriction: bool = False
    @staticmethod
    def mpi_output_all_ranks() -> None:
        Console._limit_mpi_output_ranks = None
    @staticmethod
    def mpi_output_root_rank_only(root_rank: int = 0, /, bypass_when_debug = False) -> None:
        if not isinstance(root_rank, int):
            try:
                root_rank = int(root_rank)
            except:
                raise TypeError(f"Argument provided for parameter root_rank (instance of {type(root_rank)}) is not an integer and cannot be interpreted as one.")
        if root_rank < 0:
            raise ValueError("Negative values cannot be MPI rank indexes.")
        Console.__limit_mpi_output_ranks = (root_rank, )
        Console.__debug_bypasses_mpi_output_restriction = bypass_when_debug
    @staticmethod
    def mpi_output_ranks_only(*ranks: int, bypass_when_debug = False) -> None:
        parsed_ranks = []
        for rank in ranks:
            if not isinstance(rank, int):
                try:
                    rank = int(rank)
                except:
                    raise TypeError(f"Argument provided for parameter ranks (instance of {type(rank)}) is not an integer and cannot be interpreted as one.")
            if rank < 0:
                raise ValueError("Negative values cannot be MPI rank indexes.")
            parsed_ranks.append(rank)
        Console.__limit_mpi_output_ranks = tuple(parsed_ranks)
        Console.__debug_bypasses_mpi_output_restriction = bypass_when_debug

    __process_id: str|int|None = None
    @staticmethod
    def clear_process_id() -> None:
        """
        Clear the manually set process ID.

        MPI ranks are handled automatically.
        """
        Console.__process_id = None
    @staticmethod
    def set_process_id(id: str|int) -> None:
        """
        Set process ID.

        MPI ranks are handled automatically.
        """
        Console.__process_id = id

    @staticmethod
    def print_raw(*args: str, **kwargs) -> None:
        Console.__print("", args, show_insert = False, **kwargs)
    @staticmethod
    def print_raw_verbose(*args: str, verbosity_level: int = 0, **kwargs) -> None:
        if _settings_object.verbose and verbosity_level >= _settings_object.verbosity_level:
            Console.__print("", args, show_insert = False, **kwargs)
    @staticmethod
    def print_raw_verbose_debug(*args: str, verbosity_level: int|None = -1, **kwargs) -> None:
        if _settings_object.debug and (verbosity_level is None or _settings_object.verbose and verbosity_level >= _settings_object.verbosity_level):
            Console.__print("", args, show_insert = False, **kwargs)
    @staticmethod
    def print_raw_debug(*args: str, **kwargs) -> None:
        Console.print_raw_verbose_debug(*args, verbosity_level = None, **kwargs)

    @staticmethod
    def print(*args: str, **kwargs) -> None:
        Console.__print("", args, show_insert = True, **kwargs)

    @staticmethod
    def print_verbose(*args: str, verbosity_level: int = 0, **kwargs) -> None:
        if _settings_object.verbose and verbosity_level >= _settings_object.verbosity_level:
            Console.__print("", args, show_insert = True, **kwargs)

    @staticmethod
    def print_info(*args: str, **kwargs) -> None:
        Console.__print("|INFO|", args, show_insert = True, **kwargs)
    @staticmethod
    def print_verbose_info(*args: str, verbosity_level: int = 0, **kwargs) -> None:
        if _settings_object.verbose and verbosity_level >= _settings_object.verbosity_level:
            Console.print_info(*args, **kwargs)

    @staticmethod
    def print_warning(*args: str, **kwargs) -> None:
        Console.__print("\u00BFWARN?", args, show_insert = True, **kwargs)
    @staticmethod
    def print_verbose_warning(*args: str, verbosity_level: int = 0, **kwargs) -> None:
        if _settings_object.verbose and verbosity_level >= _settings_object.verbosity_level:
            Console.print_warning(*args, **kwargs)

    @staticmethod
    def print_error(*args: str, **kwargs) -> None:
        Console.__print("!ERRO!", args, show_insert = True, **kwargs)
    @staticmethod
    def print_verbose_error(*args: str, verbosity_level: int = 0, **kwargs) -> None:
        if _settings_object.verbose and verbosity_level >= _settings_object.verbosity_level:
            Console.print_error(*args, **kwargs)

    @staticmethod
    def print_verbose_debug(*args: str, verbosity_level: int|None = -1, **kwargs) -> None:
        if _settings_object.debug and (verbosity_level is None or _settings_object.verbose and verbosity_level >= _settings_object.verbosity_level):
            Console.__print("<DBUG>", args, show_insert = True, **kwargs)
    @staticmethod
    def print_debug(*args: str, **kwargs) -> None:
        Console.print_verbose_debug(*args, verbosity_level = None, **kwargs)

    @staticmethod
    def custom_print(prefix: str, *args: str, show_insert: bool = True, **kwargs) -> None:
        if len(prefix) < 6:
            prefix = (" " * (6 - len(prefix))) + prefix
        elif len(prefix) > 6:
            prefix = prefix[:6]
        Console.__print(prefix, args, show_insert = show_insert, **kwargs)
    @staticmethod
    def custom_print_verbose(prefix: str, *args: str, show_insert: bool = True, verbosity_level: int = -1, **kwargs) -> None:
        if _settings_object.verbose and verbosity_level >= _settings_object.verbosity_level:
            Console.custom_print(prefix, *args, show_insert = show_insert, **kwargs)
    @staticmethod
    def custom_print_verbose_debug(prefix: str, *args: str, show_insert: bool = True, verbosity_level: int|None = -1, **kwargs) -> None:
        if _settings_object.debug and (verbosity_level is None or _settings_object.verbose and verbosity_level >= _settings_object.verbosity_level):
            Console.custom_print(prefix, *args, show_insert = show_insert, **kwargs)
    def custom_print_debug(prefix: str, *args: str, show_insert: bool = True, **kwargs) -> None:
        Console.custom_print_verbose_debug(prefix, *args, show_insert = show_insert, verbosity_level = None, **kwargs)

    @staticmethod
    def pause(exit = False):
        input("Press enter to {}... ".format("exit" if exit else "continue"))

    @staticmethod
    def read_line() -> str:
        return input()

    @staticmethod
    def input(prompt: str, **kwargs) -> str:
        Console.__print("|INPT|", [prompt], show_insert = False, end = "\n       >>> ", **kwargs)
        return Console.read_line()





from functools import wraps

class _PrivateModuleFunctions(object):
    @staticmethod
    def _slurm_wrapper(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            if _settings_object.slurm:
                if "flush" not in kwargs:
                    kwargs["flush"] = True
            return func(*args, **kwargs)
        return wrapper

    @staticmethod
    def _debug_wrapper(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            if _settings_object.debug:
                if "flush" not in kwargs:
                    kwargs["flush"] = True
            return func(*args, **kwargs)
        return wrapper

    @staticmethod
    def _mpi_wrapper(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            if "flush" not in kwargs:
                kwargs["flush"] = True
            if Console._limit_mpi_output_ranks is None or get_mpi_rank() in Console._limit_mpi_output_ranks:
                return func(f" ({get_mpi_rank()})", *args, **kwargs)
            else:
                return lambda *args, **kwargs: None
        if _settings_object.mpi_avalible:
            return wrapper
        else:
            return lambda *args, **kwargs: func("", *args, **kwargs)

[DeprecationWarning]
class OldConsole(object):

    __print_custom_newline_spaces = " " * 16 # "                "
    @staticmethod
    def __print_custom_newline_format(mpi_rank_insert, s):
        return str(s).replace("\n", f"\n{Console.__print_custom_newline_spaces}" + (" " * len(mpi_rank_insert)) + (" " * len(Console.__create_locak_rank_insert())))
    
    _limit_mpi_output_ranks: tuple[int]|None = None
    @staticmethod
    def mpi_output_all_ranks() -> None:
        Console._limit_mpi_output_ranks = None
    @staticmethod
    def mpi_output_root_rank_only(root_rank: int = 0) -> None:
        if not isinstance(root_rank, int):
            try:
                root_rank = int(root_rank)
            except:
                raise TypeError(f"Argument provided for parameter root_rank (instance of {type(root_rank)}) is not an integer and cannot be interpreted as one.")
        if root_rank < 0:
            raise ValueError("Negative values cannot be MPI rank indexes.")
        Console._limit_mpi_output_ranks = (root_rank, )
    @staticmethod
    def mpi_output_ranks_only(*ranks: int) -> None:
        parsed_ranks = []
        for rank in ranks:
            if not isinstance(rank, int):
                try:
                    rank = int(rank)
                except:
                    raise TypeError(f"Argument provided for parameter root_rank (instance of {type(rank)}) is not an integer and cannot be interpreted as one.")
            if rank < 0:
                raise ValueError("Negative values cannot be MPI rank indexes.")
            parsed_ranks.append(rank)
        Console._limit_mpi_output_ranks = tuple(parsed_ranks)

    __local_parallel_rank: int|None = None
    @staticmethod
    def clear_local_parallel_rank() -> None:
        """
        Remove printing of non-MPI process ranks.
        """
        Console.__local_parallel_rank = None
    @staticmethod
    def set_local_parallel_rank(rank_index: int) -> None:
        """
        Set this processes rank for the given MPI rank.

        (i.e. when there are multiple local processes spawned from the same MPI rank process)
        """
        Console.__local_parallel_rank = rank_index
    @staticmethod
    def __create_locak_rank_insert() -> str:
        return f"[{Console.__local_parallel_rank}]" if Console.__local_parallel_rank is not None else ""

    @staticmethod
    @_PrivateModuleFunctions._mpi_wrapper # Still necessary to set auto-flushing etc.
    @_PrivateModuleFunctions._debug_wrapper
    @_PrivateModuleFunctions._slurm_wrapper
    def print_raw(_, firstValue = "", *args, **kwargs):
        """
        Print with no extra layout alterations but retaining all the functionality that the Console methods offer for MPI rank limitations and auto-flushing.
        """
        print(firstValue, *args, **kwargs)

    @staticmethod
    @_PrivateModuleFunctions._mpi_wrapper # Still necessary to set auto-flushing etc.
    @_PrivateModuleFunctions._debug_wrapper
    @_PrivateModuleFunctions._slurm_wrapper
    def print_raw_verbose(_, firstValue = "", *args, **kwargs):
        """
        Print (when verbose output is enabled) with no extra layout alterations but retaining all the functionality that the Console methods offer for MPI rank limitations and auto-flushing.
        """
        if _settings_object.verbose:
            print(firstValue, *args, **kwargs)

    @staticmethod
    @_PrivateModuleFunctions._mpi_wrapper # Still necessary to set auto-flushing etc.
    @_PrivateModuleFunctions._debug_wrapper
    @_PrivateModuleFunctions._slurm_wrapper
    def print_raw_debug(_, firstValue = "", *args, **kwargs):
        """
        Print (when verbose output is enabled) with no extra layout alterations but retaining all the functionality that the Console methods offer for MPI rank limitations and auto-flushing.
        """
        if _settings_object.debug:
            print(firstValue, *args, **kwargs)

    @staticmethod
    @_PrivateModuleFunctions._mpi_wrapper
    @_PrivateModuleFunctions._debug_wrapper
    @_PrivateModuleFunctions._slurm_wrapper
    def print(mpi_rank_insert, firstValue = "", *args, **kwargs):
        rank_inserts = f"{mpi_rank_insert}{Console.__create_locak_rank_insert()}  "
        if rank_inserts == " " * 2:
            rank_inserts = ""
        print(f"{rank_inserts}{Console.__print_custom_newline_format(mpi_rank_insert, firstValue)}", *[Console.__print_custom_newline_format(mpi_rank_insert, arg) for arg in args], **kwargs)

    @staticmethod
    @_PrivateModuleFunctions._mpi_wrapper
    @_PrivateModuleFunctions._debug_wrapper
    @_PrivateModuleFunctions._slurm_wrapper
    def print_info(mpi_rank_insert, firstValue = "", *args, **kwargs):
        print(f"--|| INFO ||--{mpi_rank_insert}{Console.__create_locak_rank_insert()}  {Console.__print_custom_newline_format(mpi_rank_insert, firstValue)}", *[Console.__print_custom_newline_format(mpi_rank_insert, arg) for arg in args], **kwargs)

    def print_verbose_info(firstValue = "", *args, **kwargs):
        if _settings_object.verbose:
            Console.print_info(firstValue, *args, **kwargs)

    @staticmethod
    @_PrivateModuleFunctions._mpi_wrapper
    @_PrivateModuleFunctions._debug_wrapper
    @_PrivateModuleFunctions._slurm_wrapper
    def print_warning(mpi_rank_insert, firstValue = "", *args, **kwargs):
        print(f"--\u00BF\u00BF WARN ??--{mpi_rank_insert}{Console.__create_locak_rank_insert()}  {Console.__print_custom_newline_format(mpi_rank_insert, firstValue)}", *[Console.__print_custom_newline_format(mpi_rank_insert, arg) for arg in args], **kwargs)

    @staticmethod
    def print_verbose_warning(firstValue = "", *args, **kwargs):
        if _settings_object.verbose:
            Console.print_warning(firstValue, *args, **kwargs)

    @staticmethod
    @_PrivateModuleFunctions._mpi_wrapper
    @_PrivateModuleFunctions._debug_wrapper
    @_PrivateModuleFunctions._slurm_wrapper
    def print_error(mpi_rank_insert, firstValue = "", *args, **kwargs):
        print(f"--!! ERRO !!--{mpi_rank_insert}{Console.__create_locak_rank_insert()}  {Console.__print_custom_newline_format(mpi_rank_insert, firstValue)}", *[Console.__print_custom_newline_format(mpi_rank_insert, arg) for arg in args], **kwargs)

    def print_verbose_error(firstValue = "", *args, **kwargs):
        if _settings_object.verbose:
            Console.print_error(firstValue, *args, **kwargs)

    @staticmethod
    @_PrivateModuleFunctions._mpi_wrapper
    @_PrivateModuleFunctions._debug_wrapper
    @_PrivateModuleFunctions._slurm_wrapper
    def print_debug(mpi_rank_insert, firstValue = "", *args, **kwargs):
        if _settings_object.debug:
            print(f"--<< DEBG >>--{mpi_rank_insert}{Console.__create_locak_rank_insert()}  {Console.__print_custom_newline_format(mpi_rank_insert, firstValue)}", *[Console.__print_custom_newline_format(mpi_rank_insert, arg) for arg in args], **kwargs)

    @staticmethod
    def pause(exit = False):
        input("Press enter to {}... ".format("exit" if exit else "continue"))



from typing import cast
import warnings

from QuasarCode._global_settings import settings_object as __settings_object
from .._MPI_Config_Base import _MPI_Config_Base

if __settings_object.mpi_avalible:
    warnings.warn("Attempted to load MPI configuration type when mpi4py module is avalible.")



class _MPI_Config(_MPI_Config_Base[None]):
    """
    MPI Configuration

    (MPI disabled version)

    (readonly) comm

    (readonly) comm_size

    (readonly) rank

    (set; get) root

    (readonly) is_root

    check_is_root(int|None)

    update(comm-type, int|None)
    """

    def __init__(self, comm: None) -> None:
        super().__init__(
            comm = None,
            size = 1,
            rank = 0,
            root = 0
        )

if not _MPI_Config._is_singleton_avalible():
    _MPI_Config(None)

mpi_config: _MPI_Config = cast(_MPI_Config, _MPI_Config._get_singleton())



from copy import copy
import pytest

from QuasarCode.IO.Text import Console
from QuasarCode.IO.Configurations import PropertiesConfig, JsonConfig, YamlConfig
from QuasarCode.IO.Checkpointing import CheckpointController, PickleCheckpoint

class TestClass_IO_Text(object):
    def test_Text(self):
        assert False
        self.fail("Not Implemented - Sort Input")#TODO: sort inputs
        
        try:
            Console.pause()
            Console.pause(exit = True)
            self.assertTrue(True)#TODO: is this needed for success to be registered?
        except:
            self.fail("The output must be checked.")

class TestClass_IO_Configurations(object):
    def test_a_test(self):
        d = Ex()
        d.add_item("inner", Ex())
        d.inner.add_item("value", "Hello World!")
        print(d.inner.value)
        d.inner.value = 1
        print(d.inner.value)
        
        d.writable = False

        d.inner.value = 2
        print(d.inner.value)
        return






    def test_ConfigBase(self):
        cfg = PropertiesConfig.from_file("testconfig.properties")
        print(cfg)

        assert cfg.writable == False
        print(cfg.value1)
        assert cfg.value1 == "a string of text"
        print(cfg["value1"])
        assert cfg["value1"] == "a string of text"
        cfg.value1 = 0
        print(cfg.value1)
        assert cfg.value1 == "a string of text", "Value was changed inside non-writable object."

    def test_ConfigBase_writable(self):
        cfg = PropertiesConfig.from_file("testconfig.properties", writable = True)
        print(cfg)

        assert cfg.writable == True
        print(cfg.value1)
        assert cfg.value1 == "a string of text"
        print(cfg["value1"])
        assert cfg["value1"] == "a string of text"
        cfg.value1 = 0
        print(cfg.value1)
        assert cfg.value1 == 0

    def test_PropertiesConfig(self):
        cfg = PropertiesConfig.from_file("testconfig.properties")
        print(cfg)

        print(cfg.value1)
        assert cfg.value1 == "a string of text"
        print(cfg.value2)
        assert cfg.value2 == "100"
        print(cfg.value3)
        assert cfg.value3 == "more text"

    def test_JsonConfig(self):
        cfg = JsonConfig.from_file("testconfig.json", True, -1)
        print(cfg)

        print(cfg.value1)
        assert cfg.value1 == "a string of text"
        print(cfg.value2)
        assert cfg.value2 == 100
        print(cfg.value3)
        assert cfg.value3 == "more text"

        print(cfg.value4)
        print(cfg.value4.sub_value1)
        assert cfg.value4.sub_value1 == "a nested value"
        print(cfg.value4.sub_value2)
        assert cfg.value4.sub_value2 == 2.8

    def test_YamlConfig(self):
        cfg = YamlConfig.from_file("testconfig.yaml", True, True, -1)
        print(cfg)

        print(cfg.value1)
        assert cfg.value1 == "a string of text"
        print(cfg.value2)
        assert cfg.value2 == 100
        print(cfg.value3)
        assert cfg.value3 == "more text"

        print(cfg.value4)
        print(cfg.value4.sub_value1)
        assert cfg.value4.sub_value1 == "a nested value"
        print(cfg.value4.sub_value2_longer)
        assert cfg.value4.sub_value2_longer == 2.8

        print(cfg.value4.value4_1.sub_sub_value1)
        assert cfg.value4.value4_1.sub_sub_value1 == "double nest test"
    
    def test_YamlConfigDisplay(self):
        cfg = YamlConfig.from_file("testconfig_2.yaml", True, True, -1)
        print(cfg)
        no_indent = ""
        indent_spaces = "    "
        assert cfg.__str__() == f"""\
files:
{indent_spaces}
    input:
        simulation_snapshots = "filepath"
        snapshot_file        = "file.txt"
{indent_spaces}{indent_spaces}
        zero_metalicity_filters:
{indent_spaces}{indent_spaces}{indent_spaces}
            high_mass_halos:
                field_name    = "gas.last_halo_masses"
                min           = None
                include_lower = True
                max           = None
                include_upper = True
{indent_spaces}
    output:
{indent_spaces}{indent_spaces}
        raw_data:
            prefix                         = "SpecWizard_data"
            sightline_locations_prefix     = "rays"
            sightline_velocity_info_prefix = "ray_velocity_information"
            optical_depth_prefix           = "optical_depth"
{indent_spaces}{indent_spaces}
        hdf5:
            file_name = "data"
{indent_spaces}{indent_spaces}
        image:
            file_prefix = "SpecWizard_pixel_optical_depths"
            file_type   = "png"
{indent_spaces}{indent_spaces}
        graph:
            file_name = "SpecWizard_ray_absorbtion"
            file_type = "png"
{no_indent}
ray_properties:
    axis                       = "z"
    spacing_mpc                = 1.0
    depth_fraction_start       = 0.0
    depth_fraction_end         = 1.0
    velocity_element_depth_kms = 1.0
{no_indent}
elements:
{indent_spaces}
    hydrogen_1:
        element_name = "Hydrogen"
        ion_symbol   = "H I"
{indent_spaces}
    carbon_3:
        element_name = "Carbon"
        ion_symbol   = "C III"
{indent_spaces}
    carbon_4:
        element_name = "Carbon"
        ion_symbol   = "C IV"
{indent_spaces}
    oxygen_6:
        element_name = "Oxygen"
        ion_symbol   = "O VI"
{indent_spaces}
    silicon_3:
        element_name = "Silicon"
        ion_symbol   = "Si III"
{indent_spaces}
    silicon_4:
        element_name = "Silicon"
        ion_symbol   = "Si IV"
{indent_spaces}
    magnesium:
        element_name = "Magnesium"
        ion_symbol   = "MgII"
{no_indent}
parallelization:
    run_parallel = False
    cpus         = 64
"""



class TestClass_IO_Checkpointing(object):
    def test_PickleCheckpoint(self):
        controller_0 = CheckpointController([], "./testcheckpointfile", PickleCheckpoint)
        for i in range(0, 5):
            controller_0.data.append(i**2)
        controller_0.checkpoint()
        del controller_0

        controller_1 = CheckpointController([], "./testcheckpointfile", PickleCheckpoint)
        for i in range(5, 7):
            controller_1.data.append(i**2)
        controller_1.checkpoint()
        del controller_1

        controller_2 = CheckpointController([], "./testcheckpointfile", PickleCheckpoint)
        for i in range(7, 9):
            controller_2.data.append(i**2)
        controller_2.checkpoint()
        del controller_2

        controller_final = CheckpointController([], "./testcheckpointfile", PickleCheckpoint)
        assert controller_final.state.checkpoint_index == 2
        assert len(controller_final.data) == 9

        for i in range(9, 11):
            controller_final.data.append(i**2)

        Console.print_info(controller_final.data)

        controller_final.remove_checkpoint()
        controller_final.remove_backup()




class Ex(object):
    def __init__(self):
        self.__data = {}
        self.writable = True

    def add_item(self, key: str, value):
        self.__data[key] = value

    def __getitem__(self, *keys):
        if keys[0] in self.__data:
            value = self.__data[keys[0]]
            if len(keys) > 1:
                if isinstance(value, Ex):
                    return value[keys[1:]] if self.writable else copy(value[keys[1:]])
                else:
                    raise TypeError(f"Value of {keys[0]} is a root node and cannot be automatically indexed further.")
            else:
                return value if self.writable else copy(value)
        else:
            raise AttributeError(f"{keys[0]} is not a valid attribute or config name.")

    def __getattr__(self, key: str):
        return self[key]

    def __setattr__(self, key: str, value):
        super().__setattr__(key, value)
        #if key != "_Mapping__data" and key in self.__data:
        #    self[key] = value


from ..Tools._Struct import Struct
from ..Tools._autoproperty import NullableTypedAutoProperty
from ..Tools._typeshield import TypeShield
from ._PersonInfomation import PersonInfomation

class AuthorInfomation(Struct):
    given_name    = NullableTypedAutoProperty[str](TypeShield[str](str))
    family_name   = NullableTypedAutoProperty[str](TypeShield[str](str))
    email_address = NullableTypedAutoProperty[str](TypeShield[str](str))
    website_url   = NullableTypedAutoProperty[str](TypeShield[str](str))
    #TODO: update the type checks for email addresses and URLs

    @staticmethod
    def from_person(person: PersonInfomation):
        return AuthorInfomation(
            given_name = person.given_name,
            family_name = person.family_name,
            email_address = person.email_addresses[0],
            website_url = person.website_url
        )



from ._UndersubscribedEventError import UndersubscribedEventError
from ._SubscriberNotPresentError import SubscriberNotPresentError
from ._NotEnoughArgumentsError import NotEnoughArgumentsError

class Event(object):
    """
    Implements Event Driven programming in Python (edp)
    """

    def __init__(self):
        self.__subscribers = []
    
    def subscribe(self, func):
        """
        Registers a method with the event instance
        For syntactic sugar, use event += func
        """
        self.__subscribers.append(func)
    
    def __iadd__(self, func):
        """
        Registers a method with the event instance
        Sintactic sugar for Event.subscribe(func)
        """
        self.subscribe(func)
        return self
    
    def unsubscribe(self, func):
        """
        Removes the registration of a method from the event instance
        For syntactic sugar, use event -= func
        """
        if len(self.__subscribers) == 0:
            raise UndersubscribedEventError("No subscribers.")
        try:
            self.__subscribers.remove(func)
        except ValueError:
            raise SubscriberNotPresentError("The provided subscriber was not listed as a subscriber to the event.")
    
    def __isub__(self, func):
        """
        Removes the registration of a method from the event instance
        Sintactic sugar for Event.unsubscribe(func)
        """
        self.unsubscribe(func)
        return self
    
    def run(self, *args, **kwargs):
        """
        Calls the event passing any *args and **kwargs accross
        """
        if len(self.__subscribers) > 0:
            try:
                return [func(*args, **kwargs) for func in self.__subscribers]
            except TypeError as e:
                raise NotEnoughArgumentsError(str(e))
        else:
            raise UndersubscribedEventError("No subscribers.")

    def getTotalSubscribers(self):
        return len(self.__subscribers)


class SubscriberNotPresentError(BaseException):
    """
    Indicates that a specific subscriber is not listed as a subscriber to the event.
    """
    def __init__(self, message):
        self.message = message


import os
import sys
from typing import List

def get_source_file_directory(source_file_path: str) -> str:
    """
    Provides the absolute directory path of the source file.

    Paramiters:
        str source_file_path -> The value of the __file__ variable.

    Returns:
        str -> The path to the folder containing the current source file.
    """
    return source_file_path.rsplit(os.sep, 1)[0]

def add_to_path(directory_path: str):
    sys.path.append(os.path.abspath(directory_path))

def relitive_add_to_path(root_directory_path: str, *relitive_path_segments: List[str]):
    """
    Adds to the Python system path, a directory specified relitive to a root directory.

    Paramiters:
        str root_directory_path -> The path onto which to add the relitive path segments.
        List[str] (args) relitive_path_segments -> Relitive path or path segments to append to the root path.
    """
    add_to_path(os.path.join(root_directory_path, *relitive_path_segments))

def source_file_relitive_add_to_path(source_file_path: str, *relitive_path_segments: List[str]):
    """
    Adds to the Python system path, a directory specified relitive to the current source file.

    Paramiters:
        str source_file_path -> The value of the __file__ variable.
    """
    relitive_add_to_path(get_source_file_directory(source_file_path), *relitive_path_segments)



import numpy as np
from typing import Union, List, Tuple, Dict

from QuasarCode.Tools import AutoProperty, TypedAutoProperty, TypeCastAutoProperty, TypeShield, NestedTypeShield, Cast, NestedCast, cast_ndarray_float64, cast_dict, cast_int, cast_float, cast_list

class Test_AutoProperty(object):
    def test_standard(self):

        class T(object):
            a = AutoProperty()
            b = AutoProperty(allow_uninitialised = True)

        t = T()

        try:
            _ = t.a
            raise AssertionError("Attempt to access uninitialised property succeeded.")
        except ValueError:
            pass

        t.a = 1

        assert t.a == 1, "Failed to read expected value of property."
        assert t.b is None, "Failed to retrive uninitialised value of None."

        t.b = 2
        assert t.b == 2, "Failed to read expected value of property."

        del t.a
        del t.b

        try:
            _ = t.a
            raise AssertionError("Attempt to access uninitialised (deleted) property succeeded.")
        except ValueError:
            pass
        assert t.b is None, "Failed to retrive uninitialised (deleted) value of None."

        t.a = 10
        t.b = 20

        assert t.a == 10, "Failed to read expected value of property (re-initialised after deletion)."
        assert t.b == 20, "Failed to read expected value of property (re-initialised after deletion)."

    def test_type_check(self):

        class T(object):
            a = TypedAutoProperty[int](TypeShield[int](int, np.int8, np.int16, np.int32, np.int64))
            b = TypedAutoProperty[Tuple[List[Union[int, float, bool]]]](NestedTypeShield[Tuple[List[Union[int, float, bool]]]]([tuple], [list], [int, float, bool]))

        t = T()

        #TODO: move test to TypeShield tests!
        assert str(NestedTypeShield[Tuple[List[Union[int, float, bool]]]]([tuple], [list], [int, float, bool])) == "TypeShield(tuple -> list -> {int, float, bool})", "Failed to generate correct type string."

        try:
            _ = t.a
            raise AssertionError("Attempt to access uninitialised property succeeded.")
        except ValueError:
            pass

        try:
            _ = t.b
            raise AssertionError("Attempt to access uninitialised property succeeded.")
        except ValueError:
            pass

        t.a = 1
        t.b = ([1, 2.5, True], [False, 1.8, -100])

        assert t.a == 1, "Failed to read expected value of property."
        assert t.b == ([1, 2.5, True], [False, 1.8, -100]), "Failed to read expected value of property."

        try:
            t.a = "2"
            raise AssertionError("Attempt to access uninitialised property succeeded.")
        except TypeError:
            pass

        try:
            t.b = (["1", 2.5, True], [False, 1.8, -100])
            raise AssertionError("Attempt to assign value with inappropriate left hand branch type succeeded.")
        except TypeError:
            pass

        assert t.a == 1, "Failed to read expected value of property."
        assert t.b == ([1, 2.5, True], [False, 1.8, -100]), "Failed to read expected value of property."

        try:
            t.b = ([1, "2.5", True], "[False, 1.8, -100]")
        except TypeError:
            raise AssertionError("Attempt to assign value with inappropriate type but valid left hand branch failed.\nTHIS IS UNEXPECTED BEHAVIOUR AS ONLY LHS IS CHECKED!!!")

        
        assert t.b == ([1, "2.5", True], "[False, 1.8, -100]"), "Failed to read expected value of property."

    def test_type_cast(self):

        class T(object):
            a = TypeCastAutoProperty[np.ndarray](cast_ndarray_float64)
            b = TypeCastAutoProperty[str](Cast[str](str, nullable = True))
            c = TypeCastAutoProperty[List[Tuple[float, ...]]](NestedCast[List[Tuple[float, ...]]](list, tuple, float))
            d = TypeCastAutoProperty[List[np.ndarray]](NestedCast[List[np.ndarray]](list, cast_ndarray_float64))
            e = TypeCastAutoProperty[np.ndarray](NestedCast[np.ndarray](cast_ndarray_float64))
            f = TypeCastAutoProperty[Dict[int, float]](cast_dict(cast_int, cast_float))
            g = TypeCastAutoProperty[Dict[int, List[np.ndarray]]](cast_dict(cast_int, NestedCast[List[np.ndarray]](cast_list, cast_ndarray_float64)))

        t = T()

        try:
            _ = t.a
            raise AssertionError("Attempt to access uninitialised property succeeded.")
        except ValueError:
            pass

        try:
            _ = t.c
            raise AssertionError("Attempt to access uninitialised property succeeded.")
        except ValueError:
            pass

        try:
            _ = t.g
            raise AssertionError("Attempt to access uninitialised property succeeded.")
        except ValueError:
            pass

        t.a = [[1, 2, 3], [4, 5, 6]]
        t.c = np.array([[1, 2, 3], [4, 5, 6], [7, 8, 9]], dtype = int)
        t.d = [[1, 2, 3], [4, 5, 6], [7, 8, 9]]
        t.e = [[1, 2, 3], [4, 5, 6], [7, 8, 9]]
        t.f = { "0": 0, "1": 1, "2": 2, "3": 3, "4": 4, "5": 5 }
        t.g = { "0": [], "1": [(1,)], "2": [(2, 2), (2, 2)], "3": [(3, 3, 3), (3, 3, 3), (3, 3, 3)], "4": [(4, 4, 4, 4), (4, 4, 4, 4), (4, 4, 4, 4), (4, 4, 4, 4)], "5": [(5, 5, 5, 5, 5), (5, 5, 5, 5, 5), (5, 5, 5, 5, 5), (5, 5, 5, 5, 5), (5, 5, 5, 5, 5)] }

        assert (t.a == np.array([[1.0, 2.0, 3.0], [4.0, 5.0, 6.0]], dtype = np.float64)).all(), "Failed to read expected value of property."
        assert t.b is None, "Failed to retrive uninitialised value of None."
        assert t.c == [(1.0, 2.0, 3.0),
                       (4.0, 5.0, 6.0),
                       (7.0, 8.0, 9.0)], "Failed to read expected value of property."
        assert (t.d[0] == np.array([1, 2, 3], dtype = np.float64)).all(), "Failed to read expected value of property."
        assert (t.d[1] == np.array([4, 5, 6], dtype = np.float64)).all(), "Failed to read expected value of property."
        assert (t.d[2] == np.array([7, 8, 9], dtype = np.float64)).all(), "Failed to read expected value of property."
        assert isinstance(t.d, list)
        assert (t.e == np.array([[1, 2, 3],
                                 [4, 5, 6],
                                 [7, 8, 9]], dtype = np.float64)).all(), "Failed to read expected value of property."
        assert t.f == { 0: 0.0, 1: 1.0, 2: 2.0, 3: 3.0, 4: 4.0, 5: 5.0 }, "Failed to retrive uninitialised value of None."

        assert np.array([np.array([(t.g[i] == np.array([i for _ in range(i)], dtype = np.float64)).all() for arr_i in range(len(t.g[i]))], dtype = bool).all() for i in t.g.keys()], dtype = bool).all()
#        assert t.g == { 0: [],
#                        1: [np.array([1], dtype = np.float64)],
#                        2: [np.array([2, 2], dtype = np.float64), np.array([2, 2], dtype = np.float64)],
#                        3: [np.array([3, 3, 3], dtype = np.float64), np.array([3, 3, 3], dtype = np.float64), np.array([3, 3, 3], dtype = np.float64)],
#                        4: [np.array([4, 4, 4, 4], dtype = np.float64), np.array([4, 4, 4, 4], dtype = np.float64), np.array([4, 4, 4, 4], dtype = np.float64), np.array([4, 4, 4, 4], dtype = np.float64)],
#                        5: [np.array([5, 5, 5, 5, 5], dtype = np.float64), np.array([5, 5, 5, 5, 5], dtype = np.float64), np.array([5, 5, 5, 5, 5], dtype = np.float64), np.array([5, 5, 5, 5, 5], dtype = np.float64), np.array([5, 5, 5, 5, 5], dtype = np.float64)] }, "Failed to retrive uninitialised value of None."

        t.b = 2
        assert t.b == "2", "Failed to read expected value of property."


    def test_load(self):

        class T(object):
            a = TypeCastAutoProperty[str](NestedCast[List[np.ndarray]](list, cast_ndarray_float64))

        t = T()

        t.a = np.array([np.linspace(-100, 100, 100000) for _ in range(10000)], dtype = np.float32)

        print(t.a)



from ._CheckpointBase import CheckpointBase
from ._PickleCheckpoint import PickleCheckpoint
from ._CheckpointState import CheckpointState

class CheckpointController(object):
    def __init__(self, empty_data_structure: object, filepath: str, checkpointType: type = PickleCheckpoint, **kwargs):
        if CheckpointBase not in checkpointType.mro():
            raise TypeError("Checkpoint type idoes not inherit from the class CheckpointBase.")

        self.__checkpoint_handeler = checkpointType(filepath, **kwargs)
        self.__current_state = CheckpointState()
        
        if self.__checkpoint_handeler.checkpoint_avalible:
            self.__current_state.initialise(from_checkpoint = self.__checkpoint_handeler.read_checkpoint())

        else:
            self.__current_state.initialise()
            self.__current_state.checkpoint_data = empty_data_structure

    @property
    def data(self):
        return self.__current_state.checkpoint_data

    @data.setter
    def data(self, value):
        self.__current_state.checkpoint_data = value

    @property
    def state(self):
        return self.__current_state

    def checkpoint(self):
        self.__current_state.increment_index()
        self.__checkpoint_handeler.write_checkpoint(self.__current_state.serialise())

    def restore_backup(self):
        if not self.__checkpoint_handeler.backup_avalible:
            raise FileNotFoundError("No backup to restore from.")
        self.__current_state = CheckpointState()
        self.__current_state.initialise(from_checkpoint = self.__checkpoint_handeler.read_backup())

    def remove_checkpoint(self):
        self.__checkpoint_handeler.remove_checkpoint()

    def remove_backup(self):
        self.__checkpoint_handeler.remove_backup()



import unittest

import QuasarCode.edp as edp

class TestClass_edp(object):

    def test_standaloneMethod(self):
        def testMethod():
            return 0

        event = edp.Event()
        event += testMethod

        assert event.run()[0] == 0



class NotEnoughArgumentsError(BaseException):
    """
    Indicates that not enough arguments were passed for the function signiture of a subscribed method during event running.
    """
    def __init__(self, message):
        self.message = message